import logging
from http import HTTPStatus
from typing import Dict, Union, List

from marshmallow import Schema
from marshmallow.fields import Field

from restit.internal.response_status_parameter import ResponseStatusParameter

LOGGER = logging.getLogger(__name__)


# noinspection PyShadowingBuiltins
def response_status(status: Union[int, HTTPStatus], content_types: Dict[str, Union[Schema, Field]], description: str):
    def decorator(func):
        http_status_code = status if isinstance(status, int) or status is None else status.value
        response_status_parameter = ResponseStatusParameter(http_status_code, description, content_types)

        registered_response_status_parameters: List[ResponseStatusParameter] = \
            getattr(func, "__response_status_parameters__", [])
        LOGGER.debug(
            "Registering response status parameter %s for %s", response_status_parameter, func.__name__
        )
        registered_response_status_parameters.append(response_status_parameter)
        setattr(func, "__response_status_parameters__", registered_response_status_parameters)
        return func

    return decorator



import unittest

from marshmallow import Schema, fields, ValidationError


class MarshmallowLearningTestCase(unittest.TestCase):
    def test_dump(self):
        class MySchema(Schema):
            f1 = fields.Integer()
            f2 = fields.String()

        my_schema = MySchema()

        self.assertEqual({'f1': 2, 'f2': '3'}, my_schema.dump({"f1": "2", "f2": "3"}))
        self.assertEqual('{"f1": 2}', my_schema.dumps({"f1": "2"}))

    def test_load(self):
        class MySchema(Schema):
            f1 = fields.Integer()
            f2 = fields.String()

        my_schema = MySchema()

        self.assertEqual({'f1': 2, 'f2': '3'}, my_schema.load({"f1": "2", "f2": "3"}))
        self.assertEqual({'f1': 2, "f2": "3"}, my_schema.loads('{"f1": "2", "f2": "3"}'))

    def test_validate(self):
        class MySchema(Schema):
            f1 = fields.Integer()
            f2 = fields.String()

        my_schema = MySchema()
        self.assertEqual({}, my_schema.validate({'f1': 2, 'f2': '3'}))
        self.assertEqual({}, my_schema.validate({'f1': '2', 'f2': '3'}))
        self.assertEqual({'f1': ['Not a valid integer.']}, my_schema.validate({'f1': 'asd', 'f2': '3'}))

    def test_single_field(self):
        field = fields.Integer()

        self.assertEqual(2, field.serialize("", None, lambda *args: "2"))
        self.assertEqual(2, field.serialize("", None, lambda *args: 2))
        self.assertEqual(2, field.deserialize("2"))
        self.assertEqual(2, field.deserialize(2))

        with self.assertRaises(ValidationError):
            field.deserialize("21321dsadsad31")

        with self.assertRaises(ValueError):
            field.serialize("", None, lambda *args: "31sadsad")



import unittest

from werkzeug.exceptions import MethodNotAllowed

from restit import Hyperlink
from restit.request import Request
from restit.request_mapping_decorator import request_mapping
from restit.resource import Resource
from restit.response import Response
from restit.restit_app import RestitApp
from restit.restit_test_app import RestitTestApp


@request_mapping("/")
class MyResource(Resource):
    def get(self, request: Request, **kwargs) -> Response:
        return Response(request.get_request_body_as_type(dict))

    def post(self, request: Request, **kwargs) -> Response:
        return Response(request.get_request_body_as_type(dict), 201)

    def put(self, request: Request, **kwargs) -> Response:
        return Response(request.get_request_body_as_type(dict), 201)

    def delete(self, request: Request, **kwargs) -> Response:
        return Response(request.get_request_body_as_type(dict), 201)


@request_mapping("/no_methods")
class NoMethodsResource(Resource):
    pass


@request_mapping("/miau/:id")
class ResourceWithPathParams(Resource):
    def get(self, request: Request, **path_params) -> Response:
        return Response(path_params)


@request_mapping("/resource_with_hyperlink")
class ResourceWithHyperLink(Resource):
    def get(self, request: Request, **kwargs) -> Response:
        return Response({
            "hyperlink_with_path_params": Hyperlink(ResourceWithPathParams).generate(request, id=10),
            "hyperlink": Hyperlink(MyResource).generate(request)
        })


@request_mapping("/pass_headers")
class PassHeadersResource(Resource):
    def get(self, request: Request, **kwargs) -> Response:
        headers = request.get_headers()
        return Response(
            {
                'Accept': headers["Accept"],
                'Accept-Encoding': headers["Accept-Encoding"],
                'Content-Type': headers["Content-Type"],
                'Content-Encoding': headers["Content-Encoding"],
                'Accept-Charset': headers["Accept-Charset"]
            })


class RestitTestAppTestCase(unittest.TestCase):
    def setUp(self) -> None:
        resit_app = RestitApp(resources=[
            MyResource(),
            NoMethodsResource(),
            PassHeadersResource(),
            ResourceWithHyperLink(),
            ResourceWithPathParams()
        ])
        self.resit_test_app = RestitTestApp(resit_app)

    def test_get_json_body(self):
        response = self.resit_test_app.get("/", json={"key": "value"})
        self.assertEqual(200, response.get_status_code())
        self.assertEqual({"key": "value"}, response.json())
        self.assertEqual('{"key": "value"}', response.text)
        self.assertEqual(b'{"key": "value"}', response.content)
        self.assertEqual({
            'Content-Encoding': 'utf-8',
            'Content-Length': 16,
            'Content-Type': 'application/json'
        }, response._headers)

    def test_get_data_body(self):
        response = self.resit_test_app.get("/", data={"key": "value", "key2": "value2"})
        self.assertEqual(200, response.get_status_code())
        self.assertEqual({'key': 'value', 'key2': 'value2'}, response.json())
        self.assertEqual('{"key": "value", "key2": "value2"}', response.text)
        self.assertEqual(b'{"key": "value", "key2": "value2"}', response.content)
        self.assertEqual({
            'Content-Encoding': 'utf-8',
            'Content-Length': 34,
            'Content-Type': 'application/json'
        }, response._headers)

    def test_post(self):
        response = self.resit_test_app.post("/", json={"key": "value"})
        self.assertEqual(201, response.get_status_code())
        self.assertEqual({'key': 'value'}, response.json())
        self.assertEqual('{"key": "value"}', response.text)
        self.assertEqual(b'{"key": "value"}', response.content)
        self.assertEqual({
            'Content-Encoding': 'utf-8',
            'Content-Length': 16,
            'Content-Type': 'application/json'
        }, response.get_headers())

    def test_put(self):
        response = self.resit_test_app.put("/", json={"key": "value"})
        self.assertEqual(201, response.get_status_code())
        self.assertEqual({'key': 'value'}, response.json())
        self.assertEqual('{"key": "value"}', response.text)
        self.assertEqual(b'{"key": "value"}', response.content)
        self.assertEqual({
            'Content-Encoding': 'utf-8',
            'Content-Length': 16,
            'Content-Type': 'application/json'
        }, response._headers)

    def test_delete(self):
        response = self.resit_test_app.delete("/", json={"key": "value"})
        self.assertEqual(201, response.get_status_code())
        self.assertEqual({'key': 'value'}, response.json())
        self.assertEqual('{"key": "value"}', response.text)
        self.assertEqual(b'{"key": "value"}', response.content)
        self.assertEqual({
            'Content-Encoding': 'utf-8',
            'Content-Length': 16,
            'Content-Type': 'application/json'
        }, response._headers)

    def test_no_method(self):
        self.assertEqual(405, self.resit_test_app.get("/no_methods").get_status_code())
        self.assertEqual(405, self.resit_test_app.post("/no_methods").get_status_code())
        self.assertEqual(405, self.resit_test_app.put("/no_methods").get_status_code())
        self.assertEqual(405, self.resit_test_app.delete("/no_methods").get_status_code())
        self.assertEqual(405, self.resit_test_app.patch("/no_methods").get_status_code())
        self.assertEqual(204, self.resit_test_app.options("/no_methods").get_status_code())

    def test_raise_if_enabled(self):
        self.resit_test_app.set_raise_on_exceptions(True)
        with self.assertRaises(MethodNotAllowed):
            self.assertEqual(405, self.resit_test_app.get("/no_methods").get_status_code())
            self.assertEqual(405, self.resit_test_app.post("/no_methods").get_status_code())
            self.assertEqual(405, self.resit_test_app.put("/no_methods").get_status_code())
            self.assertEqual(405, self.resit_test_app.delete("/no_methods").get_status_code())
            self.assertEqual(405, self.resit_test_app.patch("/no_methods").get_status_code())

    def test_pass_headers(self):
        response = self.resit_test_app.get("/pass_headers", headers={"Accept-Charset": "utf-8"})
        self.assertEqual(200, response.get_status_code())
        self.assertEqual({
            'Accept': '*/*',
            'Accept-Charset': 'utf-8',
            'Accept-Encoding': 'gzip, deflate',
            'Content-Encoding': 'gzip, deflate',
            'Content-Type': 'application/octet-stream'
        }, response.json())

    def test_hyperlinks(self):
        response = self.resit_test_app.get("/resource_with_hyperlink")

        self.assertEqual(200, response.get_status_code())
        self.assertEqual({
            'hyperlink': 'http://127.0.0.1/',
            'hyperlink_with_path_params': 'http://127.0.0.1/miau/10'
        }, response.json())



import logging
from typing import Type, List

from restit.path_parameter import PathParameter
from restit.path_parameter_decorator import register_path_parameter
from restit.resource import Resource

LOGGER = logging.getLogger(__name__)


def request_mapping(path: str, path_parameters: List[PathParameter] = None):
    path_parameters = path_parameters or []

    def wrapper(clazz: Type[Resource]):

        if not path.startswith("/"):
            raise PathIsNotStartingWithSlashException(path)
        LOGGER.debug("Registering path %s for swagger %s", path, clazz.__name__)
        clazz.__request_mapping__ = path
        for path_parameter in path_parameters:
            register_path_parameter(path_parameter, clazz)
        return clazz

    return wrapper


class PathIsNotStartingWithSlashException(Exception):
    pass



from restit.open_api.open_api_documentation import OpenApiDocumentation
from .development_server import DevelopmentServer
from .hyperlink import Hyperlink
from .namespace import Namespace
from .path_parameter_decorator import path_parameter
from .query_parameter_decorator import query_parameter
from .request import Request
from .request_body_decorator import request_body
from .request_mapping_decorator import request_mapping
from .resource import Resource
from .response import Response
from .restit_app import RestitApp
from .restit_test_app import RestitTestApp
from .static_directory_resource import StaticDirectoryResource
from .static_file_response import StaticFileResponse

__version__ = "0.1.0"



from marshmallow import fields
from marshmallow.fields import Field


class PathParameter:
    _PYTHON_TYPE_FIELD_MAPPING = {
        int: fields.Integer(),
        str: fields.String()
    }

    def __init__(self, name: str, description: str, field_type: Field):
        self.name = name
        self.description = description
        self.field_type = \
            field_type if isinstance(field_type, Field) else PathParameter._PYTHON_TYPE_FIELD_MAPPING[field_type]

    def __str__(self):
        return f"PathParameter(name='{self.name}', description='{self.description}', field_type={self.field_type})"



from restit.request import Request
from restit.request_mapping_decorator import request_mapping
from restit.resource import Resource
from restit.response import Response


@request_mapping("/favicon.ico")
class DefaultFaviconResource(Resource):
    def get(self, request: Request, **path_params) -> Response:
        return Response("")



class ContactObject:
    def __init__(self, name: str, url: str, email: str):
        self.name = name
        self.url = url
        self.email = email

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "url": self.url,
            "email": self.email
        }



import unittest
from unittest.mock import patch, mock_open

from restit import RestitTestApp, RestitApp, Resource, Response, Request, request_mapping
from restit.static_file_response import StaticFileResponse


@request_mapping("/static-response")
class MyResource(Resource):
    def get(self, request: Request, **path_params) -> Response:
        file_path = request.get_request_body_as_type(dict)["file_path"]
        return StaticFileResponse(file_path)


class StaticFileResourceTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.rest_test_app = RestitTestApp(RestitApp(
            resources=[MyResource()]
        ))

    def test_text_file(self):
        m = mock_open(read_data=b"Huhu from file")
        with patch("restit.static_file_response.open", m):
            response = self.rest_test_app.get("/static-response?", json={"file_path": "/some/file/path.txt"})
            self.assertEqual(200, response.get_status_code())
            self.assertEqual("text/plain", response.get_headers()["Content-Type"])
            self.assertEqual("Huhu from file", response.text)

    def test_html_file(self):
        m = mock_open(read_data=b"<title>dummy html</title>")
        with patch("restit.static_file_response.open", m):
            response = self.rest_test_app.get("/static-response?", json={"file_path": "/some/file/path.html"})
            self.assertEqual(200, response.get_status_code())
            self.assertEqual("text/html", response.get_headers()["Content-Type"])
            self.assertEqual("<title>dummy html</title>", response.text)



import unittest

from marshmallow import Schema, fields

from restit.open_api.open_api_schema_converter import OpenApiSchemaConverter


class SimpleSchema(Schema):
    """My simple schema."""
    field1 = fields.Integer(required=True, default=1)
    field1.__doc__ = "Doc of field1"
    field2 = fields.String()
    field3 = fields.UUID()


class SchemaWithNested(Schema):
    """I have a nested field"""
    nested_field = fields.Nested(SimpleSchema(), required=True)


class SchemaWithIntegerList(Schema):
    int_list = fields.List(fields.Integer())


class SchemaWithNestedList(Schema):
    nested_list = fields.Nested(SchemaWithNested())


class OpenApiSchemaConverterTestCase(unittest.TestCase):
    def test_simple(self):
        open_api_schema = OpenApiSchemaConverter.convert_schema(SimpleSchema())

        self.assertEqual({
            'description': 'My simple schema.',
            'properties': {
                'field1': {
                    'default': 1,
                    'description': 'Doc of field1',
                    'type': 'integer'
                },
                'field2': {
                    'description': 'A string field.',
                    'type': 'string'
                },
                'field3': {
                    'description': 'A UUID field.',
                    'format': 'uuid',
                    'type': 'string'
                }
            },
            'required': ['field1'],
            'type': 'object'
        }, open_api_schema)

    def test_nested(self):
        open_api_schema = OpenApiSchemaConverter.convert_schema(SchemaWithNested())

        self.assertEqual({
            'description': 'I have a nested field',
            'properties': {
                'nested_field': {
                    'description': 'My simple schema.',
                    'properties': {
                        'field1': {
                            'default': 1,
                            'description': 'Doc '
                                           'of '
                                           'field1',
                            'type': 'integer'
                        },
                        'field2': {
                            'description': 'A '
                                           'string '
                                           'field.',
                            'type': 'string'
                        },
                        'field3': {
                            'description': 'A '
                                           'UUID '
                                           'field.',
                            'format': 'uuid',
                            'type': 'string'
                        }
                    },
                    'required': ['field1'],
                    'type': 'object'
                }
            },
            'required': ['nested_field'],
            'type': 'object'
        }, open_api_schema)

    def test_integer_list(self):
        open_api_schema = OpenApiSchemaConverter.convert(SchemaWithIntegerList())
        self.assertEqual({
            'description': None,
            'properties': {
                'int_list': {
                    'items': {
                        'description': 'An integer field.',
                        'type': 'integer'
                    },
                    'type': 'array'
                }
            },
            'required': [],
            'type': 'object'
        }, open_api_schema)

    def test_nested_list(self):
        open_api_schema = OpenApiSchemaConverter.convert(SchemaWithNestedList())

        self.assertEqual({
            'description': None,
            'properties': {
                'nested_list': {
                    'description': 'I have a nested field',
                    'properties': {
                        'nested_field': {
                            'description': 'My '
                                           'simple '
                                           'schema.',
                            'properties': {
                                'field1': {
                                    'default': 1,
                                    'description': 'Doc '
                                                   'of '
                                                   'field1',
                                    'type': 'integer'
                                },
                                'field2': {
                                    'description': 'A '
                                                   'string '
                                                   'field.',
                                    'type': 'string'
                                },
                                'field3': {
                                    'description': 'A '
                                                   'UUID '
                                                   'field.',
                                    'format': 'uuid',
                                    'type': 'string'
                                }
                            },
                            'required': ['field1'],
                            'type': 'object'
                        }
                    },
                    'required': ['nested_field'],
                    'type': 'object'
                }
            },
            'required': [],
            'type': 'object'
        }, open_api_schema)



import unittest

from restit.internal.request_deserializer_service import RequestDeserializerService


class RequestDeserializerTestCase(unittest.TestCase):
    def tearDown(self) -> None:
        RequestDeserializerService.restore_default_request_deserializers()
        RequestDeserializerService.clear_all_request_deserializers()



import unittest
from typing import List, Union

from restit.resource import Resource
from restit.restit_app import RestitApp


class BaseTestServerTestCase(unittest.TestCase):
    test_server = None
    thread = None
    resources: List[Resource] = []
    restit_app: Union[RestitApp, None]

    @classmethod
    def setUpClass(cls) -> None:
        cls.restit_app = RestitApp(resources=BaseTestServerTestCase.resources, debug=True)
        cls.port = cls.restit_app.start_development_server(port=0, blocking=False)

    @classmethod
    def tearDownClass(cls) -> None:
        cls.restit_app.stop_development_server()



import unittest

from restit import RestitApp, OpenApiDocumentation, Namespace
from restit.open_api.info_object import InfoObject
from test.restit.acceptence_test.example_resource import TodosResource
from test.restit.acceptence_test.todo_repo import TodoRepo


class RestitAppWithDevSeverTestCase(unittest.TestCase):
    def setUp(self) -> None:
        repo = TodoRepo()

        todo_namespace = Namespace("/todos", resources=[TodosResource(repo)])

        self.restit_app = RestitApp(
            namespaces=[todo_namespace],
            open_api_documentation=OpenApiDocumentation(
                info=InfoObject(title="Sample TODO Api", version="1.0.0", description="Demonstrate the RestIt library"),
                path="/todos/api"
            ), debug=True
        )

    def test_generate_spec(self):
        self.restit_app._init()

        self.restit_app.start_development_server()

        spec = self.restit_app._open_api_documentation.generate_spec()

        self.assertEqual({}, spec)



from typing import List, Tuple, Union

from restit.common import get_default_encoding, guess_text_content_subtype_string
from restit.internal.default_response_serializer.default_str_text_response_serializer import \
    DefaultStrTextResponseSerializer
from restit.internal.response_status_parameter import ResponseStatusParameter
from restit.response_serializer import ResponseSerializer


class StringFallbackResponseSerializer(ResponseSerializer):
    def get_media_type_strings(self) -> List[str]:
        return ["*/*"]

    def get_response_data_type(self) -> type:
        return str

    def validate_and_serialize(
            self, response_input: str, response_status_parameter: Union[None, ResponseStatusParameter]
    ) -> Tuple[bytes, str]:
        content_type = guess_text_content_subtype_string(response_input)
        response_input_bytes = response_input.encode(encoding=get_default_encoding())
        if self.find_schema(content_type, response_status_parameter):
            raise DefaultStrTextResponseSerializer.SchemaNotSupportedForStringResponseBody()

        return response_input_bytes, content_type



from typing import Any, List, Type, Union


class RequestDeserializer:
    def get_content_type_list(self) -> Union[List[str], None]:
        raise NotImplemented()

    def deserialize(self, request_input: bytes, encoding: str = None) -> Any:
        raise NotImplemented()

    def get_deserialized_python_type(self) -> Type:
        raise NotImplemented()

    def can_handle(self, content_type: str, python_type: Type) -> bool:
        return \
            (self.get_content_type_list() is None or content_type in self.get_content_type_list()) \
            and python_type == self.get_deserialized_python_type()



import json

from werkzeug.exceptions import HTTPException


class Rfc7807HttpProblem(Exception):
    def __init__(self, title: str, detail: str, status: int, instance: str = None, type: str = "about:blank"):
        self.title = title
        self.detail = detail
        self.status = status
        self.instance = instance
        self.type = type

    @staticmethod
    def from_http_exception(http_exception: HTTPException) -> "Rfc7807HttpProblem":
        return Rfc7807HttpProblem(
            title=http_exception.name,
            detail=http_exception.description,
            status=http_exception.code
        )

    def to_json(self) -> dict:
        return {
            "title": self.title,
            "detail": self.detail,
            "status": self.status,
            "instance": self.instance,
            "type": self.type
        }

    def __str__(self):
        return json.dumps(self.to_json())



import unittest


def my_decorator(some_argument: str):
    def decorator(func):
        setattr(func, "some_argument", some_argument)

        def wrapper(*args, **kwargs):
            return_value = func(*args, **kwargs)
            return some_argument + " " + return_value

        return wrapper

    return decorator


@my_decorator(some_argument="Hello")
def my_function(what: str):
    return what


class DecoratorLearningTestCase(unittest.TestCase):
    def test_mix_decorator_types_not_possible(self):
        self.assertIsNone(getattr(my_function, "some_argument", None))
        self.assertEqual("Hello cat", my_function("cat"))



import re
from functools import lru_cache
from typing import List, Tuple, Union, Match

from restit.internal.request_body_properties import RequestBodyProperties
from restit.internal.response_status_parameter import ResponseStatusParameter
from restit.open_api.info_object import InfoObject
from restit.open_api.open_api_schema_converter import \
    OpenApiSchemaConverter
from restit.resource import PathParameter
from restit.resource import Resource


class OpenApiDocumentation:
    _IGNORE_RESOURCE_CLASS_NAMES = ["DefaultFaviconResource", "OpenApiResource"]
    _SPEC_VERSION = "3.0.0"

    def __init__(self, info: InfoObject, path: str = "/api"):
        self.info = info
        self.path = path
        self._resources: List[Resource] = []
        self._servers = []

    def register_resource(self, resource: Resource):
        if resource not in self._resources and \
                resource.__class__.__name__ not in OpenApiDocumentation._IGNORE_RESOURCE_CLASS_NAMES:
            self._resources.append(resource)

    @lru_cache(maxsize=1)
    def generate_spec(self) -> dict:
        self._resources.sort(key=lambda r: r.__request_mapping__)
        spec_structure = self._generate_base_spec_structure()
        self._generate_paths(spec_structure)
        return spec_structure

    def _generate_paths(self, spec_structure):
        paths = spec_structure["paths"]
        for resource in self._resources:
            if resource.__request_mapping__:
                self._add_resource(paths, resource, spec_structure)

    def _add_resource(self, paths: dict, resource: Resource, spec_structure: dict):
        path, inferred_path_parameters = \
            self._infer_path_params_and_open_api_path_syntax(resource.__request_mapping__)
        paths[path] = {}
        for method_name, method_object in self._get_allowed_resource_methods(resource).items():
            paths[path][method_name] = {
                "responses": {},
                "parameters": []
            }

            method_spec = paths[path][method_name]
            self._add_summary_and_description_to_method(method_spec, method_object)
            self._add_path_parameters(method_spec, resource, inferred_path_parameters)
            self._add_query_parameters(method_spec, method_object)
            self._add_request_body(method_spec, method_object, spec_structure)
            self._add_responses(method_spec, method_object, spec_structure)

    @staticmethod
    def _add_responses(method_spec: dict, method_object: object, spec_structure: dict):
        response_status_parameters = getattr(method_object, "__response_status_parameters__", None)
        if response_status_parameters:
            for response_status_parameter in response_status_parameters:  # type: ResponseStatusParameter
                method_spec["responses"][response_status_parameter.status or "default"] = {
                    "description": response_status_parameter.description,
                    "content": OpenApiDocumentation._create_content_field(response_status_parameter.content_types)
                }

    @staticmethod
    def _add_request_body(method_spec: dict, method_object: object, spec_structure: dict):
        request_body_parameter = getattr(
            method_object, "__request_body_properties__", None)  # type: RequestBodyProperties
        if request_body_parameter:
            # ToDo allow creating global schema under components
            # spec_structure["components"]["schemas"][request_body_parameter.schema.__class__.__name__] = \
            #    MarshmallowToOpenApiSchemaConverter.convert(request_body_parameter.schema)

            method_spec["requestBody"] = {
                "description": request_body_parameter.description,
                "required": request_body_parameter.required,
                "content": OpenApiDocumentation._create_content_field(request_body_parameter.content_types)
            }

    @staticmethod
    def _create_content_field(content_types: dict):
        return {
            content_type:
                {
                    "schema": OpenApiSchemaConverter.convert(schema)
                }
            for content_type, schema in content_types.items()
        }

    @staticmethod
    def _add_query_parameters(method_spec: dict, method_object: object):
        method_spec["parameters"].extend([
            {
                "name": query_parameter.name,
                "in": "query",
                "description": query_parameter.description,
                "required": query_parameter.field_type.required,
                "schema": OpenApiSchemaConverter.convert(query_parameter.field_type)
            }
            for query_parameter in getattr(method_object, "__query_parameters__", [])
        ])

    @staticmethod
    def _add_path_parameters(method_spec: dict, resource: Resource, inferred_path_parameters: List[PathParameter]):
        parameter_definitions = getattr(resource, "__path_parameters__", [])
        path_parameters = {
            path_parameter.name: path_parameter for path_parameter in inferred_path_parameters
        }
        path_parameters.update({
            path_parameter.name: path_parameter for path_parameter in parameter_definitions
        })

        method_spec["parameters"].extend([
            {
                "name": name,
                "in": "path",
                "required": True,
                "description": path_parameter.description,
                "schema": OpenApiSchemaConverter.convert_field(path_parameter.field_type)

            } for name, path_parameter in path_parameters.items()
        ])

    @staticmethod
    def _infer_path_params_and_open_api_path_syntax(path: str) -> Tuple[str, List[PathParameter]]:
        path_parameter_list = []

        def _handle_path_parameter(match: Match) -> str:
            path_parameter_list.append(
                PathParameter(match.group(1), "", eval(match.group(2)) if match.group(2) else str)
            )
            return "{%s}" % match.group(1)

        openapi_path_syntax = re.sub(r":(\w+)(?:<(\w+)>)?", _handle_path_parameter, path)
        return openapi_path_syntax, path_parameter_list

    def _add_summary_and_description_to_method(self, method_spec: dict, method_object: object):
        summary, description = self._get_summary_and_description_from_doc(method_object.__doc__)
        method_spec["summary"] = summary
        method_spec["description"] = description

    @staticmethod
    def _get_summary_and_description_from_doc(doc: str) -> Tuple[Union[str, None], Union[str, None]]:
        if not doc:
            return None, None

        doc_split = doc.split("\n")
        summary = doc_split[0]
        description = "\n".join(doc_split[1:]) if len(doc) > 1 else None
        return summary.strip("\n "), description.strip("\n ")

    @staticmethod
    def _get_allowed_resource_methods(resource: Resource) -> dict:
        allowed_methods_dict = {allowed: getattr(resource, allowed) for allowed in resource.get_allowed_methods()}
        return allowed_methods_dict

    def _generate_base_spec_structure(self) -> dict:
        return {
            "openapi": OpenApiDocumentation._SPEC_VERSION,
            "info": self.info.to_dict(),
            "paths": {},
            "components": {
                "schemas": {}
            }
        }



import unittest
from wsgiref.util import setup_testing_defaults

import requests

from test.start_server_with_wsgi_app import start_server_with_wsgi_app


def wsgi_app(environ, start_response):
    setup_testing_defaults(environ)

    status = '200 OK'
    headers = [('Content-type', 'text/plain; charset=utf-8')]

    start_response(status, headers)

    ret = [
        f"{key}: {value}\n".encode("utf-8")
        for key, value in environ.items()
    ]
    return ret


class WsgiLearningTestCase(unittest.TestCase):
    def test_wsgi_application_get(self):
        with start_server_with_wsgi_app(wsgi_app) as port:
            response = requests.get(f"http://127.0.0.1:{port}")
            self.assertEqual(200, response.status_code)
            self.assertIn("REQUEST_METHOD: GET", response.text)

    def test_wsgi_application_post(self):
        with start_server_with_wsgi_app(wsgi_app) as port:
            response = requests.post(f"http://127.0.0.1:{port}")
            self.assertEqual(200, response.status_code)
            self.assertIn("REQUEST_METHOD: POST", response.text)



import re

from restit.common import create_dict_from_assignment_syntax


class MIMEType:
    _REGEX = re.compile(r"^([*a-zA-Z0-9_-]+)/([+*a-zA-Z0-9_-]+)(.+)?$")

    # noinspection PyShadowingBuiltins
    def __init__(self, type: str, subtype: str, quality: float = 1.0, details: dict = None):
        self.type = type if type != "*" else None
        self.subtype = subtype if subtype != "*" else None
        self.quality = quality
        self.details = details or {}

        if "q" not in self.details and quality != 1.0:
            self.details["q"] = str(quality)

        if self.type is None and self.subtype is not None:
            raise MIMEType.MIMETypeWildcardHierarchyException(type, subtype)

    @staticmethod
    def from_string(mime_type_string: str) -> "MIMEType":
        match = MIMEType._REGEX.match(mime_type_string)
        if not match:
            raise MIMEType.MIMETypeParsingException(mime_type_string)

        details = create_dict_from_assignment_syntax(match.group(3), group_delimiter=";")

        return MIMEType(
            type=match.group(1),
            subtype=match.group(2),
            quality=float(details.get("q", 1)),
            details=details
        )

    def matches_mime_type_string(self, mime_type_string: str) -> bool:
        _type, _subtype = mime_type_string.split("/")
        matches_type = self.type is None or _type == "*" or self.type == _type
        matches_subtype = self.subtype is None or _subtype == "*" or self.subtype == _subtype

        return matches_type and matches_subtype

    def to_string(self, with_details: bool = False) -> str:
        _to_string = f"{self.type or '*'}/{self.subtype or '*'}"
        if with_details:
            _to_string += ";" + ";".join([f"{key}={value}" for key, value in self.details.items()])

        return _to_string

    def __eq__(self, other: "MIMEType") -> bool:
        return self.type == other.type and self.subtype == other.subtype and self.details == other.details

    def __gt__(self, other: "MIMEType") -> bool:
        return self.quality > other.quality

    def __str__(self):
        return f"MIMEType(" \
               f"type={self.type}, subtype={self.subtype}, quality={self.quality}, details={self.details})"

    def __repr__(self):
        return str(self)

    class MIMETypeParsingException(Exception):
        pass

    class MIMETypeWildcardHierarchyException(Exception):
        pass



from marshmallow import ValidationError
from marshmallow.fields import Field

from restit.internal.request_body_properties import RequestBodyProperties
from restit.internal.type_converter.schema_or_field_deserializer import SchemaOrFieldDeserializer
from restit.request import Request


class RequestBodySchemaDeserializer:
    @staticmethod
    def deserialize(request: Request, request_body_properties: RequestBodyProperties):
        schema_or_field = request_body_properties.get_schema_for_content_type(request.get_content_type())
        _type = str if isinstance(schema_or_field, Field) else dict
        body = request.get_request_body_as_type(_type)
        try:
            request._body_type_cache[_type] = SchemaOrFieldDeserializer.deserialize(body, schema_or_field)
        except (ValidationError, ValueError) as error:
            raise request_body_properties.validation_error_class(
                f"Request body schema deserialization failed ({str(error)})"
            )

        request.get_request_body_as_type.cache_clear()



import ast
import inspect
import logging
import re
from typing import Tuple, AnyStr, Dict, Union, List

from marshmallow import ValidationError
from werkzeug.exceptions import MethodNotAllowed, BadRequest

from restit.internal.request_body_properties import RequestBodyProperties
from restit.internal.request_body_schema_deserializer import RequestBodySchemaDeserializer
from restit.internal.resource_path import ResourcePath
from restit.internal.response_serializer_service import ResponseSerializerService
from restit.internal.response_status_parameter import ResponseStatusParameter
from restit.internal.type_converter.schema_or_field_deserializer import SchemaOrFieldDeserializer
from restit.path_parameter_decorator import PathParameter
from restit.query_parameter_decorator import QueryParameter
from restit.request import Request
from restit.response import Response

LOGGER = logging.getLogger(__name__)


class Resource:
    __request_mapping__ = None
    _METHOD_NAMES = ["get", "post", "put", "delete", "patch", "options", "trace", "connect", "head"]

    def __init__(self):
        self._resource_path = None

    def init(self):
        self._resource_path = ResourcePath(self.__request_mapping__)

    def get(self, request: Request, **path_params) -> Response:
        raise MethodNotAllowed()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def post(self, request: Request, **path_params) -> Response:
        raise MethodNotAllowed()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def put(self, request: Request, **path_params) -> Response:
        raise MethodNotAllowed()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def delete(self, request: Request, **path_params) -> Response:
        raise MethodNotAllowed()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def patch(self, request: Request, **path_params) -> Response:
        raise MethodNotAllowed()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def options(self, request: Request, **path_params) -> Response:
        """Identifying allowed request methods.

        The HTTP OPTIONS method is used to describe the communication options for the target resource.
        """
        allow = " ".join(self.get_allowed_methods()).upper()
        return Response("", 204, headers={"Allow": allow})

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def trace(self, request: Request, **path_params) -> Response:
        raise MethodNotAllowed()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def connect(self, request: Request, **path_params) -> Response:
        raise MethodNotAllowed()

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def head(self, request: Request, **path_params) -> Response:
        raise MethodNotAllowed()

    def handle_request(self, request_method: str, request: Request, path_params: Dict) -> Response:
        method_object = getattr(self, request_method.lower())
        passed_path_parameters = self._collect_and_convert_path_parameters(path_params)
        self._process_query_parameters(method_object, request)
        request = self._validate_request_body(method_object, request)
        response: Response = method_object(request, **passed_path_parameters)
        response_status_parameter = Resource._find_response_schema_by_status(response.get_status_code(), method_object)
        ResponseSerializerService.validate_and_serialize_response_body(
            response, request.get_http_accept_object(), response_status_parameter
        )
        return response

    @staticmethod
    def _find_response_schema_by_status(status: int, method_object: object) -> Union[None, ResponseStatusParameter]:
        response_status_parameters = getattr(method_object, "__response_status_parameters__", [])
        for response_status_parameter in response_status_parameters:  # type: ResponseStatusParameter
            if response_status_parameter.status == status:
                return response_status_parameter

        LOGGER.warning("Response status code %d is not expected for %s", status, method_object)

    @staticmethod
    def _validate_request_body(method_object: object, request: Request) -> Request:
        request_body_properties: RequestBodyProperties = \
            getattr(method_object, "__request_body_properties__", None)
        if request_body_properties:
            RequestBodySchemaDeserializer.deserialize(request, request_body_properties)

        return request

    @staticmethod
    def _process_query_parameters(method_object, request):
        for query_parameter in getattr(method_object, "__query_parameters__", []):  # type: QueryParameter
            value: str = request.get_query_parameters().get(query_parameter.name, )
            if value is None and query_parameter.required:
                # ToDo message
                raise BadRequest()

            value = ast.literal_eval(value) if value.startswith("[") else value
            # noinspection PyProtectedMember
            request._query_parameters[query_parameter.name] = SchemaOrFieldDeserializer.deserialize(
                value, query_parameter.field_type
            )

    def _collect_and_convert_path_parameters(self, path_params: dict):
        for path_parameter in getattr(self, "__path_parameters__", []):  # type: PathParameter
            try:
                path_parameter_value = path_params[path_parameter.name]
            except KeyError:
                raise Resource.PathParameterNotFoundException(
                    f"Unable to find {path_parameter} in incoming path parameters {path_params}"
                )
            try:
                path_params[path_parameter.name] = \
                    SchemaOrFieldDeserializer.deserialize(path_parameter_value, path_parameter.field_type)
            except ValidationError as error:
                raise BadRequest(
                    f"Path parameter value '{path_parameter_value}' is not matching '{path_parameter}' "
                    f"({str(error)})"
                )

        return path_params

    def get_allowed_methods(self) -> List[str]:
        allowed = []
        for method_name in Resource._METHOD_NAMES:
            method_object = getattr(self, method_name)
            resource_method_code = inspect.getsource(method_object)
            match = re.match(
                r"^.+def\s+\w+\(self,\s*request:\s*Request.+\)\s*->\s*Response:.+raise\s+MethodNotAllowed\(\).*$",
                resource_method_code, flags=re.DOTALL | re.IGNORECASE
            )
            if match is None:
                allowed.append(method_name)

        return allowed

    def _get_match(self, url: str) -> Tuple[bool, Union[None, Dict[str, AnyStr]]]:
        assert self._resource_path
        return self._resource_path.get_match(url)

    class PathParameterNotFoundException(Exception):
        pass



import unittest

from werkzeug.datastructures import MIMEAccept


class MimeAcceptLearningTestCase(unittest.TestCase):
    def test_something(self):
        mime_accept = MIMEAccept([("application/json", 1)])
        mime_accept2 = MIMEAccept([("application/json", 1), ("text/plain", 0.6)])

        self.assertTrue(mime_accept.accept_json)
        self.assertIn("application/json", mime_accept)

        self.assertEqual(['application/json'], list(set(mime_accept2.values()) & set(mime_accept.values())))

        self.assertEqual(["application/json"], list(set(mime_accept.values()).intersection({"application/json"})))
        self.assertEqual([], list(set(mime_accept.values()).intersection({"my/type"})))
        self.assertEqual([], list(set(mime_accept2.values()).intersection({"my/type"})))
        self.assertEqual(["application/json"], list(set(mime_accept2.values()).intersection({"application/json"})))
        self.assertEqual(["application/json"], list(set(mime_accept2.values()).intersection({"application/json"})))

        self.assertEqual("application/json", mime_accept2.best_match(["application/json", "text/plain"]))

        self.assertEqual(0.6, mime_accept2[mime_accept2.find("text/plain")][1])

        self.assertEqual("application/json", MIMEAccept([("*/*", 1)]).best_match(["application/json"]))
        self.assertIsNone(MIMEAccept([("application/json", 1)]).best_match(["text/plain"]))

        mime_accept3 = MIMEAccept([("application/json", 1), ("text/html", 1), ("text/*", 1), ("*/*", 1)])

        self.assertEqual("application/json", mime_accept3.best_match(["application/json"]))
        self.assertEqual("text/html", mime_accept3.best_match(["text/html"]))
        self.assertEqual("text/plain", mime_accept3.best_match(["text/plain"]))



import requests
from marshmallow import Schema, fields

from restit.internal.request_body_properties import RequestBodyProperties
from restit.request import Request
from restit.request_body_decorator import request_body
from restit.request_mapping_decorator import request_mapping
from restit.resource import Resource
from restit.response import Response
from test.base_test_server_test_case import BaseTestServerTestCase


class RequestBodySchema(Schema):
    param1 = fields.Int(required=True)
    param2 = fields.Str(required=True)


@request_mapping("/miau")
class QueryParametersResource(Resource):
    @request_body({"application/x-www-form-urlencoded": RequestBodySchema()}, description="Huhu")
    def post(self, request: Request) -> Response:
        return Response(request.get_request_body_as_type(dict))


class RequestBodyValidationTestCase(BaseTestServerTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        BaseTestServerTestCase.resources = [
            QueryParametersResource()
        ]
        BaseTestServerTestCase.setUpClass()

    def test_request_body_validation(self):
        response = requests.post(f"http://127.0.0.1:{self.port}/miau", data={"param1": "1", "param2": "huhu"})
        self.assertEqual(200, response.status_code)
        self.assertEqual({'param1': 1, 'param2': 'huhu'}, response.json())

    def test_request_body_validation_fails(self):
        response = requests.post(f"http://127.0.0.1:{self.port}/miau", data={"param1": "hans", "param2": 222})
        self.assertEqual(422, response.status_code)
        self.assertEqual(
            "<title>422 Unprocessable Entity</title>\n"
            "<h1>Unprocessable Entity</h1>\n"
            "<p>Request body schema deserialization failed (invalid literal for int() with base 10: 'hans')</p>\n",
            response.text
        )

    def test_request_body_schema_type_not_supported(self):
        with self.assertRaises(RequestBodyProperties.UnsupportedSchemaTypeException):
            RequestBodyProperties({"text/plain": str}, "", True)

    def test_unsupported_media_type(self):
        response = requests.post(f"http://127.0.0.1:{self.port}/miau", json={"param1": "hans", "param2": 222})
        self.assertEqual(415, response.status_code)



from io import BytesIO
from json import dumps
from typing import Union, Tuple
from urllib.parse import urlparse
from wsgiref.util import setup_testing_defaults

from restit.common import get_default_encoding
from restit.request import Request
from restit.response import Response
from restit.restit_app import RestitApp


class RestitTestApp(RestitApp):

    def __init__(self, restit_app: RestitApp):
        super().__init__(
            resources=restit_app._resources,
            namespaces=restit_app._namespaces,
            debug=restit_app._debug,
            raise_exceptions=restit_app._raise_exceptions
        )
        self._init()

    def get(self, path: str, json: dict = None, data: Union[dict, str, bytes] = None, headers: dict = None) -> Response:
        return self._get_response_for_method(path, json, data, headers, "GET")

    def post(
            self, path: str, json: dict = None, data: Union[dict, str, bytes] = None, headers: dict = None) -> Response:
        return self._get_response_for_method(path, json, data, headers, "POST")

    def put(
            self, path: str, json: dict = None, data: Union[dict, str, bytes] = None, headers: dict = None) -> Response:
        return self._get_response_for_method(path, json, data, headers, "PUT")

    def delete(
            self, path: str, json: dict = None, data: Union[dict, str, bytes] = None, headers: dict = None) -> Response:
        return self._get_response_for_method(path, json, data, headers, "DELETE")

    def patch(
            self, path: str, json: dict = None, data: Union[dict, str, bytes] = None, headers: dict = None) -> Response:
        return self._get_response_for_method(path, json, data, headers, "PATCH")

    def options(
            self, path: str, json: dict = None, data: Union[dict, str, bytes] = None, headers: dict = None) -> Response:
        return self._get_response_for_method(path, json, data, headers, "OPTIONS")

    def _get_response_for_method(
            self, path: str, json: dict, data: Union[dict, str, bytes], headers: dict, method: str):
        wsgi_environment = self._create_wsgi_environment(json, data, headers, path, method)
        response = self._get_response(wsgi_environment)
        return response

    def _get_response(self, wsgi_environment):
        request = Request(wsgi_environment)
        resource, path_params = self._find_resource_for_url(request.get_path())
        if self._raise_exceptions:
            return self._get_response_or_raise_not_found(path_params, request, resource)
        else:
            return self._create_response_and_handle_exceptions(path_params, request, resource)

    def _create_wsgi_environment(
            self, json: Union[dict, None], data: Union[dict, None], header: Union[dict, None], path: str, method: str
    ):
        header = header or {}
        wsgi_environment = {}
        parsed_path = urlparse(path)
        setup_testing_defaults(wsgi_environment)
        body_as_bytes = b""
        wsgi_environment["HTTP_ACCEPT"] = header.get("Accept", "*/*")
        wsgi_environment["HTTP_ACCEPT_CHARSET"] = header.get("Accept-Charset", get_default_encoding())
        wsgi_environment["CONTENT_ENCODING"] = header.get("Content-Encoding", "gzip, deflate")
        wsgi_environment["HTTP_ACCEPT_ENCODING"] = header.get("Accept-Encoding", "gzip, deflate")
        content_type = "application/octet-stream"
        accept_charset = header.get("Accept-Charset", get_default_encoding())
        if json is not None:
            body_as_bytes = dumps(json).encode(encoding=accept_charset)
            content_type = "application/json"
        elif data is not None:
            body_as_bytes, content_type = \
                self._get_body_as_bytes_from_data_argument(data, accept_charset)
        wsgi_environment["REQUEST_METHOD"] = method
        wsgi_environment["PATH_INFO"] = parsed_path.path
        wsgi_environment["CONTENT_LENGTH"] = len(body_as_bytes)
        wsgi_environment["wsgi.input"] = BytesIO(body_as_bytes)
        wsgi_environment["QUERY_STRING"] = parsed_path.query
        wsgi_environment["CONTENT_TYPE"] = header.get("Content-Type", content_type)
        return wsgi_environment

    @staticmethod
    def _get_body_as_bytes_from_data_argument(
            data: Union[dict, str, bytes], accept_charset: str) -> Tuple[bytes, str]:
        if isinstance(data, dict):
            body_as_bytes = "&".join([f"{key}={value}" for key, value in data.items()])
            return body_as_bytes.encode(encoding=accept_charset), "application/x-www-form-urlencoded"
        elif isinstance(data, str):
            return data.encode(encoding=accept_charset), "text/plain"
        elif isinstance(data, bytes):
            return data, "text/plain"
        else:
            raise RestitTestApp.UnsupportedDataTypeException(type(data))

    class UnsupportedDataTypeException(Exception):
        pass



import logging
import traceback
from contextlib import contextmanager
from functools import lru_cache
from typing import Iterable, Callable, List, Tuple, Dict, Union

# noinspection PyProtectedMember
from werkzeug.exceptions import HTTPException, InternalServerError, NotFound

from restit.development_server import DevelopmentServer
from restit.internal.default_favicon_resource import DefaultFaviconResource
from restit.internal.exception_response_maker import HttpExceptionResponseMaker
from restit.namespace import Namespace
from restit.open_api.open_api_documentation import OpenApiDocumentation
from restit.open_api.open_api_resource import OpenApiResource
from restit.request import Request
from restit.resource import Resource
from restit.response import Response
from restit.rfc7807_http_problem import Rfc7807HttpProblem

LOGGER = logging.getLogger(__name__)


class RestitApp:
    def __init__(
            self, resources: List[Resource] = None,
            namespaces: List[Namespace] = None,
            debug: bool = False,
            raise_exceptions: bool = False,
            open_api_documentation: OpenApiDocumentation = None

    ):
        self._debug = debug
        self._namespaces: List[Namespace] = []
        self._resources: List[Resource] = []
        self._raise_exceptions = raise_exceptions
        self._open_api_documentation = open_api_documentation
        self.register_namespaces(namespaces or [])
        self.register_resources(resources or [])

        self.__development_server: Union[DevelopmentServer, None] = None

        self._init_called = False

    def set_open_api_documentation(self, open_api_documentation: OpenApiDocumentation):
        self._open_api_documentation = open_api_documentation

    def set_raise_on_exceptions(self, raise_on_exceptions: bool):
        self._raise_exceptions = raise_on_exceptions

    def register_resources(self, resources: List[Resource]):
        self.__check_resource_request_mapping(resources)
        self._resources.extend(resources)
        self._create_url_ordered_resource()

    def register_namespaces(self, namespaces: List[Namespace]):
        for namespace in namespaces:
            self.register_resources(namespace.get_adapted_resources())

    def start_development_server(self, host: str = None, port: int = 5000, blocking: bool = True) -> int:
        self.__development_server = DevelopmentServer(self, host, port)
        self.__development_server.start(blocking=blocking)
        return self.__development_server.server.server_port

    def stop_development_server(self):
        self.__development_server.stop()

    @contextmanager
    def start_development_server_in_context(self, host: str = None, port: int = 5000) -> int:
        self.__development_server = DevelopmentServer(self, host, port)
        with self.__development_server.start_in_context() as port:
            yield port

    @staticmethod
    def __check_resource_request_mapping(resources):
        for resource in resources:
            if resource.__request_mapping__ is None:
                raise RestitApp.MissingRequestMappingException(
                    f"The resource class {resource.__class__.__name__} does not appear to have a @request_mapping(...)"
                )

    def _init(self):
        self._resources.append(DefaultFaviconResource())
        if self._open_api_documentation:
            self._resources.append(OpenApiResource(self._open_api_documentation))
        for resource in self._resources:
            resource.init()
            if self._open_api_documentation:
                self._open_api_documentation.register_resource(resource)
        self._init_called = True

    def __call__(self, wsgi_environ: dict, start_response: Callable) -> Iterable:
        if not self._init_called:
            self._init()

        request = Request(wsgi_environ)
        resource, path_params = self._find_resource_for_url(request.get_path())

        response = self._create_response_and_handle_exceptions(path_params, request, resource)

        header_as_list = [(key, str(value)) for key, value in response.get_headers().items()]
        start_response(response.get_status_string(), header_as_list)

        return [response.content]

    def _create_response_and_handle_exceptions(
            self, path_params: dict, request: Request, resource: Resource
    ) -> Response:
        try:
            response = self._get_response_or_raise_not_found(path_params, request, resource)
        except HTTPException as exception:
            rfc7807_http_problem = Rfc7807HttpProblem.from_http_exception(exception)
            response = self._create_rfc7807_response(request, rfc7807_http_problem)
        except Rfc7807HttpProblem as problem:
            response = self._create_rfc7807_response(request, problem)
        except Exception as exception:
            if self._raise_exceptions:
                raise
            LOGGER.error(str(exception))
            LOGGER.error(traceback.format_exc())
            internal_exception = InternalServerError(str(exception) if self._debug else "")
            exception_response_maker = HttpExceptionResponseMaker(
                Rfc7807HttpProblem.from_http_exception(internal_exception)
            )
            response = exception_response_maker.create_response(request.get_http_accept_object())
        return response

    @staticmethod
    def _create_rfc7807_response(request: Request, rfc7807_http_problem: Rfc7807HttpProblem) -> Response:
        LOGGER.info(str(rfc7807_http_problem))
        exception_response_maker = HttpExceptionResponseMaker(rfc7807_http_problem)
        response = exception_response_maker.create_response(request.get_http_accept_object())
        return response

    @staticmethod
    def _get_response_or_raise_not_found(path_params: dict, request: Request, resource: Resource) -> Response:
        if resource is not None:
            # noinspection PyBroadException
            # noinspection PyProtectedMember
            response = resource.handle_request(
                request_method=request.get_request_method_name(),
                request=request,
                path_params=path_params
            )

        else:
            raise NotFound()
        return response

    @lru_cache()
    def _find_resource_for_url(self, url: str) -> Union[Tuple[None, None], Tuple[Resource, Dict]]:
        for resource in self._resources:
            # noinspection PyProtectedMember
            is_matching, path_params = resource._get_match(url)
            if is_matching:
                return resource, path_params

        return None, None

    def _create_url_ordered_resource(self):
        self._resources = sorted(self._resources, key=lambda r: r.__request_mapping__, reverse=True)

    class MissingRequestMappingException(Exception):
        pass



# https://wiki.selfhtml.org/wiki/MIME-Type/%C3%9Cbersicht
SUFFIX_MEDIA_TYPE_MAPPING = {
    ".html": "text/html",
    ".css": "text/css",
    ".js": "text/javascript",
    ".txt": "text/plain",
    # A
    ".dwg": "application/acad",
    ".asd": "application/astound",
    ".asn": "application/astound",
    ".tsp": "application/dsptype",
    # image
    ".png": "image/png",
    ".gif": "image/gif",
    ".ief": "image/ief",
    ".jpeg": "image/jpeg",
    ".jpg": "image/jpeg",
    ".jpe": "image/jpeg",
    ".svg": "image/svg+xml",
    ".tif": "image/tif",
    ".tiff": "image/tif",
    ".vasa": "image/mcf",
    ".bmp": "image/bmp",
    ".cod": "image/cis-cod",
    ".ras": "image/cmu-raster",
    ".fif": "image/fif",
    ".wbmp": "image/vnd.wap.wbmp",
    ".fh4": "image/x-freehand",
    ".fh5": "image/x-freehand",
    ".fhc": "image/x-freehand",
    ".ico": "image/x-icon",
    ".pnm": "image/x-portable-anymap",
    ".pbm": "image/x-portable-bitmap",
    ".pgm": "image/x-portable-graymap",
    ".ppm": "image/x-portable-pixmap",
    ".rgb": "image/x-rgb",
    ".xwd": "image/x-windowdump",
    ".xbm": "image/x-xbitmap",
    ".xpm": "image/x-xpixmap",
}



import unittest

import requests

from restit.development_server import DevelopmentServer
from restit.request import Request
from restit.request_mapping_decorator import request_mapping
from restit.resource import Resource
from restit.response import Response
from restit.restit_app import RestitApp


@request_mapping("/")
class MyResource(Resource):
    def get(self, request: Request) -> Response:
        return Response("Hello")


class DevelopmentServerTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.restit_app = RestitApp(resources=[MyResource()])
        self.development_server = DevelopmentServer(self.restit_app)

    def test_start_in_context(self):
        with self.development_server.start_in_context() as port:
            response = requests.get(f"http://127.0.0.1:{port}/")
            self.assertEqual(200, response.status_code)
            self.assertEqual("Hello", response.text)

    def test_restit_app_in_context(self):
        with self.restit_app.start_development_server_in_context(port=0) as port:
            response = requests.get(f"http://127.0.0.1:{port}/")
            self.assertEqual(200, response.status_code)
            self.assertEqual("Hello", response.text)



from typing import Union, Type, Any

from marshmallow import Schema
from marshmallow.fields import Field


class SchemaOrFieldDeserializer:
    @staticmethod
    def deserialize(value, type_or_schema: Union[Type, Schema, Field]) -> Any:
        if isinstance(type_or_schema, Schema):
            return type_or_schema.dump(value)
        if isinstance(type_or_schema, Field):
            return type_or_schema.deserialize(value)
        else:
            raise SchemaOrFieldDeserializer.OnlySchemaOrFieldTypeSupportedException()

    class OnlySchemaOrFieldTypeSupportedException(Exception):
        pass



from http import HTTPStatus
from json import loads
from typing import Union, Any

from restit.common import get_default_encoding


class Response:

    def __init__(
            self,
            response_body: Any,
            status_code: Union[int, HTTPStatus] = 200,
            headers: dict = None
    ):
        self.response_body_input = response_body
        self._status: HTTPStatus = HTTPStatus(status_code, None) if isinstance(status_code, int) else status_code
        self._headers = headers or {}
        self._headers.setdefault("Content-Encoding", get_default_encoding())
        self.content = b""
        self.text = ""

    def _prepare_headers(self, content_type: str):
        self._headers.setdefault("Content-Type", content_type)
        self._headers.setdefault("Content-Length", len(self.content))

    def get_status_string(self) -> str:
        return f"{self._status.value} {self._status.name}"

    def get_status_code(self) -> int:
        return self._status.value

    def json(self, **kwargs) -> dict:
        return loads(self.content.decode(encoding=self.get_headers()["Content-Encoding"]), **kwargs)

    def get_headers(self) -> dict:
        return self._headers

    def get_content_type(self, fallback: str = None) -> str:
        return self._headers.get("Content-Type", fallback or "application/octet-stream")

    class ResponseBodyTypeNotSupportedException(Exception):
        pass



import os
import sys
from html import escape

_DEFAULT_ENCODING = sys.getdefaultencoding()


def set_default_encoding(default_encoding: str):
    global _DEFAULT_ENCODING
    _DEFAULT_ENCODING = default_encoding


def get_default_encoding() -> str:
    global _DEFAULT_ENCODING
    return _DEFAULT_ENCODING


def get_open_api_resource_path() -> str:
    import restit
    path = os.path.join(restit.__path__[0], "resource", "swagger")
    assert os.path.isdir(path)
    return path


def create_dict_from_assignment_syntax(request_input_string: str, group_delimiter: str = "&") -> dict:
    if request_input_string is None or "=" not in request_input_string:
        return {}

    return {
        key: escape(value) for key, value in [
            pair.split("=") for pair in request_input_string.strip(group_delimiter).split(group_delimiter)
        ]
    }


def guess_text_content_subtype_bytes(content: bytes) -> str:
    if b"<html>" in content or b"<title>" in content or b"<body>" in content:
        return "text/html"

    return "text/plain"


def guess_text_content_subtype_string(content: str) -> str:
    if "<html>" in content or "<title>" in content or "<body>" in content:
        return "text/html"

    return "text/plain"



from contextlib import contextmanager
from threading import Thread
from wsgiref.simple_server import make_server


@contextmanager
def start_server_with_wsgi_app(wsgi_app):
    with make_server('', 0, wsgi_app) as httpd:
        thread = Thread(target=httpd.serve_forever)
        thread.start()
        try:
            yield httpd.server_port
        finally:
            httpd.shutdown()
            thread.join()



from http import HTTPStatus
from typing import Union, Dict, Type

from marshmallow import Schema
from marshmallow.fields import Field


class ResponseStatusParameter:
    def __init__(
            self, status: Union[int, HTTPStatus, None], description: str, content_types: Dict[str, Union[Schema, Type]]
    ):
        self.status: int = status if isinstance(status, int) or status is None else status.value
        self.description = description
        self.content_types = content_types
        self._check_request_properties_schema_type()

    def _check_request_properties_schema_type(self):
        for schema_or_field in self.content_types.values():
            if not isinstance(schema_or_field, (Field, Schema)):
                raise ResponseStatusParameter.UnsupportedSchemaTypeException(type(schema_or_field))

    class UnsupportedSchemaTypeException(Exception):
        pass



import requests

from restit import Hyperlink
from restit.internal.default_favicon_resource import DefaultFaviconResource
from restit.request import Request
from restit.request_mapping_decorator import request_mapping
from restit.resource import Resource
from restit.response import Response
from restit.restit_app import RestitApp
from test.base_test_server_test_case import BaseTestServerTestCase


@request_mapping("/")
class MyResource(Resource):
    def get(self, request: Request, **kwargs) -> Response:
        return Response("Hallo")


@request_mapping("/no_methods")
class NoMethodsResource(Resource):
    pass


@request_mapping("/miau")
class MyResource2(Resource):
    def get(self, request: Request, **kwargs) -> Response:
        return Response("wuff")

    def post(self, request: Request, **path_params) -> Response:
        return Response("", 204)


@request_mapping("/miau/:id")
class ResourceWithPathParams(Resource):

    def get(self, request: Request, **path_params) -> Response:
        return Response(path_params)


@request_mapping("/error")
class ErrorResource(Resource):
    def get(self, request: Request, **kwargs) -> Response:
        raise Exception("OH NOOOO")


@request_mapping("/resource_with_hyperlink")
class ResourceWithHyperLink(Resource):
    def get(self, request: Request, **kwargs) -> Response:
        return Response({
            "hyperlink_with_path_params": Hyperlink(ResourceWithPathParams).generate(request, id=10),
            "hyperlink": Hyperlink(MyResource).generate(request)
        })


@request_mapping("/resource_with_hyperlink_error")
class ResourceWithHyperLinkError(Resource):
    def get(self, request: Request, **kwargs) -> Response:
        return Response({
            "hyperlink_with_path_params": Hyperlink(ResourceWithPathParams).generate(request, not_there=10),
        })


class RestitAppTestCase(BaseTestServerTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        BaseTestServerTestCase.resources = [
            MyResource(),
            MyResource2(),
            ResourceWithPathParams(),
            ErrorResource(),
            NoMethodsResource(),
            ResourceWithHyperLink(),
            ResourceWithHyperLinkError(),
            DefaultFaviconResource()
        ]
        BaseTestServerTestCase.setUpClass()

    def test_simple_get_resource(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/")
        self.assertEqual(200, response.status_code)
        self.assertEqual("Hallo", response.text)

        response = requests.get(f"http://127.0.0.1:{self.port}/miau")
        self.assertEqual(200, response.status_code)
        self.assertEqual("wuff", response.text)

    def test_method_not_allowed_405(self):
        for method in ["get", "delete", "put", "post", "patch", "trace", "connect"]:
            response = requests.request(method, f"http://127.0.0.1:{self.port}/no_methods")
            self.assertEqual(405, response.status_code)
            self.assertIn("405 Method Not Allowed", response.text)

        response = requests.head(f"http://127.0.0.1:{self.port}/no_methods")
        self.assertEqual(405, response.status_code)
        self.assertEqual("", response.text)

    def test_options_method(self):
        response = requests.options(f"http://127.0.0.1:{self.port}/miau")
        self.assertEqual(204, response.status_code)
        self.assertEqual("GET POST OPTIONS", response.headers["Allow"])

    def test_url_not_found(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/NOT_THERE")
        self.assertEqual(404, response.status_code)

    def test_resource_with_path_params(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/miau/21")
        self.assertEqual(200, response.status_code)
        self.assertEqual({"id": "21"}, response.json())

    def test_internal_server_error(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/error")
        self.assertEqual(500, response.status_code)
        self.assertEqual(
            "<title>500 Internal Server Error</title>\n"
            "<h1>Internal Server Error</h1>\n"
            "<p>OH NOOOO</p>\n",
            response.text
        )

    def test_internal_server_error_as_rfc7807_json(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/error",
                                headers={'Accept': "application/json", "Accept-Charset": "utf-8"})
        self.assertEqual(500, response.status_code)
        self.assertEqual({
            'detail': 'OH NOOOO',
            'instance': None,
            'status': 500,
            'title': 'Internal Server Error',
            'type': 'about:blank'
        }, response.json()
        )
        self.assertEqual("application/problem+json", response.headers["Content-Type"])

    def test_missing_request_mapping(self):
        class ResourceWithoutRequestMapping(Resource):
            pass

        with self.assertRaises(RestitApp.MissingRequestMappingException):
            RestitApp(resources=[ResourceWithoutRequestMapping()])

    def test_hyperlink(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/resource_with_hyperlink")
        self.assertEqual(200, response.status_code)
        self.assertEqual({
            "hyperlink_with_path_params": f"http://127.0.0.1:{self.port}/miau/10",
            "hyperlink": f"http://127.0.0.1:{self.port}/"
        }, response.json())

    def test_hyperlink_path_param_not_found(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/resource_with_hyperlink_error")
        self.assertEqual(500, response.status_code)
        self.assertEqual(
            "<title>500 Internal Server Error</title>\n"
            "<h1>Internal Server Error</h1>\n"
            "<p>The path parameter id in request mapping '/miau/:id' was not found in the provided path parameters "
            "{'not_there': 10}</p>\n",
            response.text
        )

    def test_default_favicon(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/favicon.ico")
        self.assertEqual(200, response.status_code)
        self.assertEqual("", response.text)



from functools import lru_cache
from typing import List

from restit.resource import Resource


class Namespace:
    def __init__(self, path: str, resources: List[Resource] = None):
        self.__path = path
        self.__resources: List[Resource] = resources or []

    def register_resources(self, resources: List[Resource]):
        self.__resources.extend(resources)

    @lru_cache(maxsize=1)
    def get_adapted_resources(self) -> List[Resource]:
        for resource in self.__resources:
            resource.__request_mapping__ = self._prepend_path(resource.__request_mapping__)

        return self.__resources

    def _prepend_path(self, url: str) -> str:
        return "/".join([self.__path.rstrip("/"), url.lstrip("/")])



import unittest
from uuid import UUID

from marshmallow import fields

from restit import RestitApp, RestitTestApp, Request
from restit.query_parameter_decorator import query_parameter
from restit.request_mapping_decorator import request_mapping
from restit.resource import Resource
from restit.response import Response


@request_mapping("/1")
class QueryParametersResource(Resource):
    @query_parameter("param1", description="First parameter", field_type=fields.Integer())
    @query_parameter("uuid", description="uuid parameter", field_type=fields.UUID())
    def get(self, request: Request) -> Response:
        assert isinstance(request.get_query_parameters()["uuid"], UUID)

        return Response(
            {
                "param1": request.get_query_parameters()["param1"],
                "uuid": str(request.get_query_parameters()["uuid"])
            }
        )


@request_mapping("/2")
class QueryParameterListResource(Resource):
    @query_parameter("int_list", description="A list of ints", field_type=fields.List(fields.Integer()))
    def get(self, request: Request, **path_params) -> Response:
        return Response(request.get_query_parameters())


class QueryParameterTest(unittest.TestCase):
    def setUp(self) -> None:
        restit_app = RestitApp(resources=[
            QueryParametersResource(),
            QueryParameterListResource()
        ])
        self.restit_test_app = RestitTestApp(restit_app)

    def test_query_parameter(self):
        response = self.restit_test_app.get("/1?param1=3&uuid=08695ead-392a-40ab-99fa-2fe64c3b48b4")
        self.assertEqual(200, response.get_status_code())
        self.assertEqual({"param1": 3, "uuid": '08695ead-392a-40ab-99fa-2fe64c3b48b4'}, response.json())

    def test_query_parameter_list(self):
        response = self.restit_test_app.get("/2?int_list=[1,2,3,4]")
        self.assertEqual(200, response.get_status_code())
        self.assertEqual({"int_list": [1, 2, 3, 4]}, response.json())



import json
from typing import List, Tuple, Union

from restit.common import get_default_encoding
from restit.internal.response_status_parameter import ResponseStatusParameter
from restit.response_serializer import ResponseSerializer


class DefaultDictJsonResponseSerializer(ResponseSerializer):
    def get_media_type_strings(self) -> List[str]:
        return ["application/json", "application/problem+json"]

    def get_response_data_type(self) -> type:
        return dict

    def validate_and_serialize(
            self, response_input: dict, response_status_parameter: Union[None, ResponseStatusParameter]
    ) -> Tuple[bytes, str]:
        content_type = "application/json"
        schema = ResponseSerializer.find_schema(content_type, response_status_parameter)
        if schema:
            json_string = schema.dumps(response_input)
        else:
            json_string = json.dumps(response_input)
        return json_string.encode(encoding=get_default_encoding()), content_type



import logging
from typing import Dict, Union

from marshmallow import Schema
from marshmallow.fields import Field
from werkzeug.exceptions import UnprocessableEntity

from restit.internal.request_body_properties import RequestBodyProperties

LOGGER = logging.getLogger(__name__)


def request_body(
        content_types: Dict[str, Union[Schema, Field]],
        description: str,
        required: bool = True,
        validation_error_class=UnprocessableEntity
):
    def decorator(func):
        request_body_properties = RequestBodyProperties(
            content_types, description, required, validation_error_class
        )
        LOGGER.debug(
            "Registering request body parameter %s for %s", request_body_properties, func.__name__
        )
        setattr(func, "__request_body_properties__", request_body_properties)
        return func

    return decorator



import unittest
from typing import List, Tuple, Any, Union

from werkzeug.exceptions import NotAcceptable

from restit.internal.default_response_serializer.default_dict_json_response_serializer import \
    DefaultDictJsonResponseSerializer
from restit.internal.default_response_serializer.default_dict_text_response_serializer import \
    DefaultDictTextResponseSerializer
from restit.internal.default_response_serializer.dict_fallback_response_serializer import DictFallbackResponseSerializer
from restit.internal.default_response_serializer.str_fallback_response_serializer import \
    StringFallbackResponseSerializer
from restit.internal.http_accept import HttpAccept
from restit.internal.mime_type import MIMEType
from restit.internal.response_serializer_service import ResponseSerializerService
from restit.internal.response_status_parameter import ResponseStatusParameter
from restit.response import Response
from restit.response_serializer import ResponseSerializer


class ResponseSerializerTestCase(unittest.TestCase):
    def tearDown(self) -> None:
        ResponseSerializerService.restore_default_response_serializer()

    def test_default_dict_to_json(self):
        response = Response({"key": "value"})
        ResponseSerializerService.validate_and_serialize_response_body(
            response, HttpAccept([MIMEType("application", "json")])
        )

        self.assertEqual(b'{"key": "value"}', response.content)

    def test_default_dict_to_text(self):
        response = Response({"key": "value"})
        ResponseSerializerService.validate_and_serialize_response_body(
            response, HttpAccept.from_accept_string("text/plain")
        )

        self.assertEqual(b'{"key": "value"}', response.content)

    def test_default_str_to_text(self):
        response = Response("Test")
        ResponseSerializerService.validate_and_serialize_response_body(
            response, HttpAccept.from_accept_string("text/plain")
        )

        self.assertEqual(b'Test', response.content)

    def test_clear_all_default_serializer(self):
        ResponseSerializerService.clear_all_response_serializer()
        response = Response("Test")
        with self.assertRaises(NotAcceptable):
            ResponseSerializerService.validate_and_serialize_response_body(
                response, HttpAccept.from_accept_string("text/plain")
            )

    def test_register_response_serializer(self):
        class MyResponseSerializer(ResponseSerializer):
            def get_response_data_type(self) -> type:
                return str

            def get_media_type_strings(self) -> List[str]:
                return ["my/type"]

            def validate_and_serialize(
                    self, response_input: Any, response_status_parameter: Union[None, ResponseStatusParameter]
            ) -> Tuple[bytes, str]:
                return "".join(reversed(response_input)).encode(), "my/type"

        ResponseSerializerService.register_response_serializer(MyResponseSerializer())
        response = Response("Test")
        ResponseSerializerService.validate_and_serialize_response_body(
            response, HttpAccept.from_accept_string("my/type")
        )

        self.assertEqual(b'tseT', response.content)

    def test_prioritize_media_type(self):
        http_accept = HttpAccept([MIMEType("application", "json"), MIMEType("text", "plain", 0.7)])

        ResponseSerializerService.clear_all_response_serializer()

        ResponseSerializerService.register_response_serializer(DefaultDictJsonResponseSerializer())
        ResponseSerializerService.register_response_serializer(DefaultDictTextResponseSerializer())

        response = Response({"key": "value"})
        ResponseSerializerService.validate_and_serialize_response_body(response, http_accept)

        self.assertEqual(b'{"key": "value"}', response.content)
        self.assertEqual("application/json", response._headers["Content-Type"])

    def test_dict_fallback_response_serializer(self):
        ResponseSerializerService.register_response_serializer(DictFallbackResponseSerializer())
        response = Response({"key": "value"})
        ResponseSerializerService.validate_and_serialize_response_body(
            response, HttpAccept.from_accept_string("wuff/miau")
        )

        self.assertEqual(b'{"key": "value"}', response.content)
        self.assertEqual("application/json", response._headers["Content-Type"])

    def test_string_fallback_response_serializer(self):
        ResponseSerializerService.register_response_serializer(StringFallbackResponseSerializer())
        response = Response("huhu")
        ResponseSerializerService.validate_and_serialize_response_body(
            response, HttpAccept.from_accept_string("wuff/miau")
        )

        self.assertEqual(b'huhu', response.content)
        self.assertEqual("text/plain", response._headers["Content-Type"])

    def test_response_body_type_not_supported(self):
        response = Response(1.0)
        with self.assertRaises(Response.ResponseBodyTypeNotSupportedException):
            ResponseSerializerService.validate_and_serialize_response_body(
                response, HttpAccept.from_accept_string("application/json")
            )



import unittest

import requests
from marshmallow import Schema, fields

from restit import Resource, Response, Request, request_mapping, RestitApp, OpenApiDocumentation, query_parameter, \
    request_body
from restit.open_api.contact_object import ContactObject
from restit.open_api.info_object import InfoObject
from restit.open_api.license_object import LicenseObject
from restit.response_status_decorator import response_status

EXPECTED_OPEN_API_DICT = {
    'openapi': '3.0.0',
    'info': {
        'title': 'First OpenApi Test',
        'version': '1.2.3',
        'description': 'Super description',
        'termsOfService': 'http://example.com/terms/',
        'contact': {
            'name': 'API Support',
            'url': 'http://www.example.com/support',
            'email': 'support@example.com'
        },
        'license': {
            'name': 'Apache 2.0',
            'url': 'https://www.apache.org/licenses/LICENSE-2.0.html'
        }
    },
    'paths': {
        '/path': {
            'get': {
                'responses': {},
                'parameters': [{
                    'name': 'param1',
                    'in': 'query',
                    'description': 'A query parameter',
                    'required': False,
                    'schema': {
                        'type': 'integer',
                        'description': 'An integer field.',
                        'default': 10
                    }
                }],
                'summary': 'This is a summary.',
                'description': 'And here we go with a description'
            },
            'post': {
                'responses': {
                    'default': {
                        'description': 'Hmm...some default',
                        'content': {
                            'text/plain': {
                                'schema': {
                                    'type': 'integer',
                                    'description': 'An integer field.',
                                    'default': 10
                                }
                            }
                        }
                    },
                    200: {
                        'description': 'Everything worked fine',
                        'content': {
                            'text/plain': {
                                'schema': {
                                    'type': 'integer',
                                    'description': 'An integer field.',
                                    'default': 10
                                }
                            }
                        }
                    }
                },
                'parameters': [],
                'summary': None,
                'description': None,
                'requestBody': {
                    'description': 'A request body',
                    'required': True,
                    'content': {
                        'application/json': {
                            'schema': {
                                'required': ['field1'],
                                'type': 'object',
                                'description': 'A bird with a flight speed exceeding that of an unladen swallow.\n    ',
                                'properties': {
                                    'field2': {
                                        'type': 'integer',
                                        'description': 'An integer field.',
                                        'default': 10
                                    },
                                    'field1': {
                                        'type': 'string',
                                        'description': 'A string field.'
                                    }
                                }
                            }
                        },
                        'image/png': {
                            'schema': {
                                'type': 'string',
                                'description': 'A string field.'
                            }
                        }
                    }
                }
            },
            'options': {
                'responses': {},
                'parameters': [],
                'summary': 'Identifying allowed request methods.',
                'description': 'The HTTP OPTIONS method is used to describe the communication options for the target resource.'
            }
        },
        '/path/{id}/wuff/{id2}': {
            'get': {
                'responses': {},
                'parameters': [{
                    'name': 'id',
                    'in': 'path',
                    'required': True,
                    'description': '',
                    'schema': {
                        'type': 'integer',
                        'description': 'An integer field.',
                        'default': 10
                    }
                }, {
                    'name': 'id2',
                    'in': 'path',
                    'required': True,
                    'description': '',
                    'schema': {
                        'type': 'string',
                        'description': 'A string field.'
                    }
                }],
                'summary': None,
                'description': None
            },
            'options': {
                'responses': {},
                'parameters': [{
                    'name': 'id',
                    'in': 'path',
                    'required': True,
                    'description': '',
                    'schema': {
                        'type': 'integer',
                        'description': 'An integer field.',
                        'default': 10
                    }
                }, {
                    'name': 'id2',
                    'in': 'path',
                    'required': True,
                    'description': '',
                    'schema': {
                        'type': 'string',
                        'description': 'A string field.'
                    }
                }],
                'summary': 'Identifying allowed request methods.',
                'description': 'The HTTP OPTIONS method is used to describe the communication options for the target resource.'
            }
        }
    },
    'components': {
        'schemas': {}
    }
}


class MyRequestBodySchema(Schema):
    """A bird with a flight speed exceeding that of an unladen swallow.
    """

    __reusable_open_api_component__ = True

    field1 = fields.String(required=True)
    field1.__doc__ = "Description for field1"
    field2 = fields.Integer()


@request_mapping("/path")
class FirstResource(Resource):
    @query_parameter("param1", description="A query parameter", field_type=fields.Integer(default=10))
    def get(self, request: Request, **path_params) -> Response:
        """This is a summary.

        And here we go with a description
        """
        return Response("Hallo")

    @request_body(
        {
            "application/json": MyRequestBodySchema(),
            "image/png": fields.String(required=True)
        }, description="A request body", required=True
    )
    @response_status(200, {"text/plain": fields.Integer()}, "Everything worked fine")
    @response_status(None, {"text/plain": fields.Integer()}, "Hmm...some default")
    def post(self, request: Request, **path_params) -> Response:
        return Response("123", 201)


@request_mapping("/path/:id<int>/wuff/:id2")
class SecondResource(Resource):
    def get(self, request: Request, **path_params) -> Response:
        return Response("Hallo")


class OpenApiSpecTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.open_api_documentation = OpenApiDocumentation(
            info=InfoObject(
                title="First OpenApi Test", description="Super description", version="1.2.3",
                contact=ContactObject("API Support", "http://www.example.com/support", "support@example.com"),
                license=LicenseObject("Apache 2.0", "https://www.apache.org/licenses/LICENSE-2.0.html"),
                terms_of_service="http://example.com/terms/"
            )
        )

        self.open_api_documentation.register_resource(FirstResource())
        self.open_api_documentation.register_resource(SecondResource())

    def test_generate_spec(self):
        self.maxDiff = None
        open_api_dict = self.open_api_documentation.generate_spec()

        self.assertEqual(EXPECTED_OPEN_API_DICT, open_api_dict)

    def test_serve_open_api(self):
        restit_app = RestitApp(
            resources=[
                FirstResource(),
                SecondResource()
            ], debug=True, raise_exceptions=True
        )

        restit_app.set_open_api_documentation(self.open_api_documentation)

        # restit_app.start_development_server()

        with restit_app.start_development_server_in_context(port=0) as port:
            response = requests.get(f"http://127.0.0.1:{port}/api/")
            self.assertEqual(200, response.status_code)
            self.assertIn("text/html", response.headers["Content-Type"])
            self.assertIn("<title>Swagger UI</title>", response.text)

            response = requests.get(f"http://127.0.0.1:{port}/api/swagger.json")
            self.assertEqual(200, response.status_code)



from typing import Type, Union, List

from restit.common import get_default_encoding
from restit.request_deserializer import RequestDeserializer


class DefaultAnyStringDeserializer(RequestDeserializer):
    def get_deserialized_python_type(self) -> Type:
        return str

    def get_content_type_list(self) -> Union[List[str], None]:
        return None

    def deserialize(self, request_input: bytes, encoding: str = None) -> str:
        return request_input.decode(encoding or get_default_encoding())



import unittest

from marshmallow import fields

from restit import request_mapping, Request, Response, path_parameter, RestitApp, RestitTestApp
from restit.path_parameter import PathParameter
from restit.resource import Resource


@request_mapping("/path/:id1/and/:id2/and/:id3")
@path_parameter("id1", field_type=fields.Integer(), description="First path parameter")
@path_parameter("id2", field_type=fields.Float(), description="Second path parameter")
@path_parameter("id3", description="Id3")
class Resource1(Resource):
    def get(self, request: Request, **path_params) -> Response:
        return Response(path_params)


@request_mapping(
    "/path/:id", path_parameters=[PathParameter("id", "Super path parameter", fields.Integer())]
)
class Resource2(Resource):
    def get(self, request: Request, **path_params) -> Response:
        return Response(path_params)


class PathParameterTestCase(unittest.TestCase):
    def setUp(self) -> None:
        restit_app = RestitApp(resources=[Resource1(), Resource2()])
        self.restit_test_app = RestitTestApp(restit_app)

    def test_path_parameters(self):
        response = self.restit_test_app.get("/path/1/and/10/and/20")
        self.assertEqual(200, response.get_status_code())
        self.assertEqual({'id1': 1, 'id2': 10.0, 'id3': '20'}, response.json())

    def test_path_parameters_alternative_declaration(self):
        response = self.restit_test_app.get("/path/1")
        self.assertEqual(200, response.get_status_code())
        self.assertEqual({'id': 1}, response.json())

    def test_conversion_exception(self):
        response = self.restit_test_app.get("/path/1/and/hans/and/20")
        self.assertEqual(400, response.get_status_code())
        self.assertEqual(
            "<title>400 Bad Request</title>\n"
            "<h1>Bad Request</h1>\n"
            "<p>Path parameter value 'hans' is not matching 'PathParameter(name='id2', "
            "description='Second path parameter', field_type=<fields.Float(default=<marshmallow.missing>, "
            "attribute=None, validate=None, required=False, load_only=False, dump_only=False, "
            "missing=<marshmallow.missing>, allow_none=False, error_messages={'required': 'Missing data for "
            "required field.', 'null': 'Field may not be null.', 'validator_failed': 'Invalid value.', 'invalid': "
            "'Not a valid number.', 'too_large': 'Number too large.', 'special': 'Special numeric values "
            "(nan or infinity) are not permitted.'})>)' (Not a valid number.)</p>\n",
            response.text
        )



import unittest

from restit.internal.mime_type import MIMEType


class MIMETypeTestCase(unittest.TestCase):
    def test_create_from_string(self):
        mime_type_string = "application/signed-exchange+xml;v=b3;q=0.9"
        mime_type = MIMEType.from_string(mime_type_string)

        self.assertEqual("application", mime_type.type)
        self.assertEqual("signed-exchange+xml", mime_type.subtype)
        self.assertEqual(0.9, mime_type.quality)
        self.assertEqual("b3", mime_type.details["v"])

    def test_create_from_string_without_details(self):
        mime_type_string = "text/html"
        mime_type = MIMEType.from_string(mime_type_string)

        self.assertEqual("text", mime_type.type)
        self.assertEqual("html", mime_type.subtype)
        self.assertEqual(1.0, mime_type.quality)
        self.assertEqual({}, mime_type.details)

    def test_parse_wildcard(self):
        mime_type = MIMEType.from_string("*/*;q=0.8")

        self.assertIsNone(mime_type.type)
        self.assertIsNone(mime_type.subtype)
        self.assertEqual(0.8, mime_type.quality)

    def test_incorrect_wildcard_order(self):
        with self.assertRaises(MIMEType.MIMETypeWildcardHierarchyException):
            MIMEType.from_string("*/html")

    def test_parsing_failed(self):
        with self.assertRaises(MIMEType.MIMETypeParsingException):
            MIMEType.from_string("can not parse me")

    def test_mime_type_equal(self):
        mime_type1 = MIMEType("application", "json", 0.9)
        mime_type2 = MIMEType("application", "json", 0.9)

        self.assertEqual(mime_type1, mime_type2)

    def test_quality_compare(self):
        mime_type1 = MIMEType("application", "json")
        mime_type2 = MIMEType("text", "html", 0.9)

        self.assertGreater(mime_type1, mime_type2)
        self.assertLess(mime_type2, mime_type1)

    def test_matches_mime_type_string(self):
        self.assertTrue(MIMEType("text", "html").matches_mime_type_string("text/html"))
        self.assertTrue(MIMEType("text", "*").matches_mime_type_string("text/html"))
        self.assertTrue(MIMEType("text", "html").matches_mime_type_string("text/*"))
        self.assertTrue(MIMEType("*", "*").matches_mime_type_string("text/*"))

        self.assertFalse(MIMEType("text", "*").matches_mime_type_string("application/json"))

    def test_to_string(self):
        mime_type = MIMEType("application", "json", 0.9, {"v": "b3"})

        self.assertEqual("application/json", mime_type.to_string(with_details=False))
        self.assertEqual("application/json;v=b3;q=0.9", mime_type.to_string(with_details=True))



from functools import lru_cache
from typing import Any, Type
from urllib.parse import quote

import werkzeug

from restit.common import create_dict_from_assignment_syntax, get_default_encoding
from restit.internal.http_accept import HttpAccept
from restit.internal.request_deserializer_service import RequestDeserializerService


class Request:
    """https://www.python.org/dev/peps/pep-0333/"""

    def __init__(self, wsgi_environment: dict):
        self._wsgi_environment = wsgi_environment
        self._content_length = int(wsgi_environment.get("CONTENT_LENGTH", 0) or 0)
        self._query_string = wsgi_environment["QUERY_STRING"]
        self._path = wsgi_environment["PATH_INFO"]
        self._method_name = wsgi_environment["REQUEST_METHOD"]

        self._query_parameters: dict = create_dict_from_assignment_syntax(self._query_string)
        self._headers = self._get_headers(wsgi_environment)
        self._body = None

        self._request_deserializer_service = RequestDeserializerService()
        self._body_type_cache = {}

    @staticmethod
    def _get_headers(wsgi_environment: dict) -> dict:
        return {
            "Accept": wsgi_environment.get("HTTP_ACCEPT", "*/*"),
            "Accept-Encoding": wsgi_environment.get("HTTP_ACCEPT_ENCODING", get_default_encoding()),
            "Content-Type": wsgi_environment["CONTENT_TYPE"],
            "Content-Encoding": wsgi_environment.get("CONTENT_ENCODING"),
            "Accept-Charset": wsgi_environment.get("HTTP_ACCEPT_CHARSET", get_default_encoding())
        }

    def _get_body_from_wsgi_environment(self, wsgi_environment: dict) -> bytes:
        return wsgi_environment["wsgi.input"].read(self._content_length)

    @lru_cache()
    def get_request_body_as_type(self, python_type: Type) -> Any:
        return self._body_type_cache.get(
            python_type,
            self._request_deserializer_service.deserialize_request_body(
                self.get_body(), self.get_content_type(), python_type,
                self.get_headers().get("Accept-Charset", get_default_encoding())
            )
        )

    def is_body_deserializable_to_type(self, python_type: Type) -> bool:
        return self._request_deserializer_service.is_body_deserializable_to_type(self.get_content_type(), python_type)

    def is_json(self) -> bool:
        return self.get_content_type().lower() == "application/json"

    def get_path(self) -> str:
        return self._path

    def get_request_method_name(self) -> str:
        return self._method_name

    def get_headers(self) -> dict:
        return self._headers

    def get_query_parameters(self) -> dict:
        return self._query_parameters

    def get_query_string(self) -> str:
        return self._query_string

    def get_encoding(self) -> str:
        return self._headers["Content-Encoding"]

    def get_content_type(self) -> str:
        return self._headers["Content-Type"]

    def get_body(self) -> bytes:
        if not self._body:
            self._body = self._get_body_from_wsgi_environment(self._wsgi_environment)
        return self._body

    def get_http_accept(self) -> str:
        return self.get_headers()["Accept"]

    @lru_cache(maxsize=1)
    def get_http_accept_object(self) -> HttpAccept:
        return HttpAccept.from_accept_string(self.get_http_accept())

    def set_body_from_string(self, body: str):
        self._body = body.encode(self.get_encoding())

    @lru_cache(maxsize=1)
    def get_werkzeug_request(self) -> werkzeug.wrappers.Request:
        return werkzeug.wrappers.Request(self._wsgi_environment)

    @lru_cache()
    def get_host_url(self) -> str:
        """https://www.python.org/dev/peps/pep-0333/#url-reconstruction"""
        url = self._wsgi_environment['wsgi.url_scheme'] + '://'

        if self._wsgi_environment.get('HTTP_HOST'):
            url += self._wsgi_environment['HTTP_HOST']
        else:
            url += self._wsgi_environment['SERVER_NAME']

            if self._wsgi_environment['wsgi.url_scheme'] == 'https':
                if self._wsgi_environment['SERVER_PORT'] != '443':
                    url += ':' + self._wsgi_environment['SERVER_PORT']
            else:
                if self._wsgi_environment['SERVER_PORT'] != '80':
                    url += ':' + self._wsgi_environment['SERVER_PORT']

        return url

    @lru_cache()
    def get_original_url(self) -> str:
        """https://www.python.org/dev/peps/pep-0333/#url-reconstruction"""
        url = self.get_host_url()

        url += quote(self._wsgi_environment.get('SCRIPT_NAME', ''))
        url += quote(self._wsgi_environment.get('PATH_INFO', ''))
        if self._wsgi_environment.get('QUERY_STRING'):
            url += '?' + self._wsgi_environment['QUERY_STRING']

        return url



from typing import Union

import marshmallow
from marshmallow import Schema, fields
from marshmallow.fields import Field, Nested

from restit.bytes_field import BytesField


class OpenApiSchemaConverter:
    _SCHEMA_TYPE_MAPPING = {
        fields.Integer: {"type": "integer"},
        fields.Float: {"type": "number", "format": "float"},
        fields.String: {"type": "string"},
        fields.UUID: {"type": "string", "format": "uuid"},
        BytesField: {"type": "string", "format": "binary"},
        fields.Boolean: {"type": "boolean"},
        fields.Email: {"type": "string", "format": "email"},
        fields.Url: {"type": "string", "format": "uri"},
        fields.DateTime: {"type": "string", "format": "date-time"},
        fields.Date: {"type": "string", "format": "date"},
        fields.Number: {"type": "number"}
    }

    @staticmethod
    def convert(schema_or_field: Union[Field, Schema]):
        if isinstance(schema_or_field, Schema):
            return OpenApiSchemaConverter.convert_schema(schema_or_field)
        elif isinstance(schema_or_field, Nested):
            return OpenApiSchemaConverter.convert_schema(schema_or_field.schema)
        else:
            return OpenApiSchemaConverter.convert_field(schema_or_field)

    @staticmethod
    def convert_schema(schema: Schema):
        open_api_schema = {
            "required": [],
            "type": "object",
            "description": schema.__doc__,
            "properties": {
            }
        }
        for name, field in schema.fields.items():
            if isinstance(field, Nested):
                open_api_schema["properties"][name] = OpenApiSchemaConverter.convert_schema(field.schema)
            elif isinstance(field, fields.List):
                open_api_schema["properties"][name] = OpenApiSchemaConverter.convert_array(field)
            else:
                open_api_schema["properties"][name] = OpenApiSchemaConverter.convert_field(field)
            if field.required:
                open_api_schema["required"].append(name)

        return open_api_schema

    @staticmethod
    def convert_array(array: fields.List) -> dict:
        list_schema = {
            "type": "array",
            "items": OpenApiSchemaConverter.convert(array.inner)
        }
        return list_schema

    @staticmethod
    def convert_field(field: Field) -> dict:
        field_schema = OpenApiSchemaConverter._SCHEMA_TYPE_MAPPING[field.__class__]
        field_schema["description"] = OpenApiSchemaConverter._get_first_doc_line(str(field.__doc__))
        if field.default != marshmallow.missing:
            field_schema["default"] = field.default
        return field_schema

    @staticmethod
    def _get_first_doc_line(doc: str) -> str:
        lines = doc.split("\n")
        if lines:
            return lines[0].strip("\n")



from typing import Type

from restit.internal.resource_path import ResourcePath
from restit.request import Request
from restit.resource import Resource


class Hyperlink:
    def __init__(self, resource_class: Type[Resource]):
        self.resource_class = resource_class

    def generate(self, request: Request, **path_parameter) -> str:
        request_mapping_with_values = \
            ResourcePath.generate_url_with_path_parameter_values(
                self.resource_class.__request_mapping__, path_parameter
            )

        return request.get_host_url().rstrip("/") + "/" + request_mapping_with_values.lstrip("/")



import json
from typing import List, Type

from restit.common import get_default_encoding
from restit.request_deserializer import RequestDeserializer


class DefaultApplicationJsonDictDeserializer(RequestDeserializer):
    def get_content_type_list(self) -> List[str]:
        return ["application/json", "application/problem+json"]

    def deserialize(self, request_input: bytes, encoding: str = None) -> dict:
        return json.loads(request_input.decode(encoding or get_default_encoding()))

    def get_deserialized_python_type(self) -> Type:
        return dict



from typing import Any, List, Tuple, Union

from marshmallow import Schema

from restit.internal.http_accept import HttpAccept
from restit.internal.response_status_parameter import ResponseStatusParameter


class ResponseSerializer:
    def __init__(self):
        self.priority = 0

    def can_handle_incoming_media_type(self, http_accept: HttpAccept) -> bool:
        best_match = http_accept.get_best_match(self.get_media_type_strings())
        if best_match is not None:
            self.priority = best_match[1].quality
            return True
        return False

    def get_media_type_strings(self) -> List[str]:
        raise NotImplemented()

    def get_response_data_type(self) -> type:
        raise NotImplemented()

    def validate_and_serialize(
            self, response_input: Any, response_status_parameter: Union[None, ResponseStatusParameter]
    ) -> Tuple[bytes, str]:
        """Returns a tuple of the serialized bytes and the content type"""
        raise NotImplemented()

    @staticmethod
    def find_schema(
            content_type: str, response_status_parameter: Union[None, ResponseStatusParameter]
    ) -> Union[None, Schema]:
        if response_status_parameter is not None:
            try:
                return response_status_parameter.content_types[content_type]
            except KeyError:
                raise ResponseSerializer.ContentTypeNotExpectedForResponseStatusException(content_type)

    class ContentTypeNotExpectedForResponseStatusException(Exception):
        pass

    class ResponseBodyValidationFailedException(Exception):
        pass



from typing import List, Tuple, Union

from restit.internal.default_response_serializer.default_dict_json_response_serializer import \
    DefaultDictJsonResponseSerializer
from restit.internal.response_status_parameter import ResponseStatusParameter
from restit.response_serializer import ResponseSerializer


class DefaultDictTextResponseSerializer(ResponseSerializer):

    def get_media_type_strings(self) -> List[str]:
        return ["text/plain"]

    def get_response_data_type(self) -> type:
        return dict

    def validate_and_serialize(
            self, response_input: dict, response_status_parameter: Union[None, ResponseStatusParameter]
    ) -> Tuple[bytes, str]:
        response_in_bytes, _ = DefaultDictJsonResponseSerializer().validate_and_serialize(
            response_input, response_status_parameter
        )

        return response_in_bytes, "text/plain"



import logging
from typing import List

from marshmallow.fields import Field, String

from restit.path_parameter import PathParameter

LOGGER = logging.getLogger(__name__)


# noinspection PyShadowingNames
def register_path_parameter(path_parameter: PathParameter, clazz):
    registered_path_parameters: List[PathParameter] = getattr(clazz, "__path_parameters__", [])
    LOGGER.debug("Registering path parameter %s for %s", path_parameter, clazz.__name__)
    registered_path_parameters.append(path_parameter)
    setattr(clazz, "__path_parameters__", registered_path_parameters)


# noinspection PyShadowingBuiltins
def path_parameter(name: str, description: str, field_type: Field = String()):
    def decorator(clazz):
        _path_parameter = PathParameter(name, description, field_type)
        register_path_parameter(_path_parameter, clazz)
        return clazz

    return decorator



from typing import List, Union

from werkzeug.exceptions import NotAcceptable

from restit.internal.default_response_serializer.default_bytes_text_response_serializer import \
    DefaultBytesTextResponseSerializer
from restit.internal.default_response_serializer.default_dict_json_response_serializer import \
    DefaultDictJsonResponseSerializer
from restit.internal.default_response_serializer.default_dict_text_response_serializer import \
    DefaultDictTextResponseSerializer
from restit.internal.default_response_serializer.default_str_text_response_serializer import \
    DefaultStrTextResponseSerializer
from restit.internal.http_accept import HttpAccept
from restit.internal.response_status_parameter import ResponseStatusParameter
from restit.response import Response
from restit.response_serializer import ResponseSerializer

_DEFAULT_RESPONSE_SERIALIZER = [
    DefaultDictJsonResponseSerializer(),
    DefaultStrTextResponseSerializer(),
    DefaultBytesTextResponseSerializer(),
    DefaultDictTextResponseSerializer(),
]


class ResponseSerializerService:
    _RESPONSE_SERIALIZER: List[ResponseSerializer] = _DEFAULT_RESPONSE_SERIALIZER.copy()

    @staticmethod
    def register_response_serializer(response_serializer: ResponseSerializer):
        ResponseSerializerService._RESPONSE_SERIALIZER.insert(0, response_serializer)

    @staticmethod
    def clear_all_response_serializer():
        ResponseSerializerService._RESPONSE_SERIALIZER = []

    @staticmethod
    def restore_default_response_serializer():
        ResponseSerializerService._RESPONSE_SERIALIZER = _DEFAULT_RESPONSE_SERIALIZER.copy()

    @staticmethod
    def get_matching_response_serializer_for_media_type(http_accept: HttpAccept) -> List[ResponseSerializer]:
        matching_response_serializer = [
            response_serializer
            for response_serializer in ResponseSerializerService._RESPONSE_SERIALIZER
            if response_serializer.can_handle_incoming_media_type(http_accept)
        ]

        return sorted(matching_response_serializer, key=lambda s: s.priority, reverse=True)

    @staticmethod
    def validate_and_serialize_response_body(
            response: Response, http_accept: HttpAccept,
            response_status_parameter: Union[None, ResponseStatusParameter] = None
    ):
        matching_response_serializer_list = ResponseSerializerService.get_matching_response_serializer_for_media_type(
            http_accept
        )
        if not matching_response_serializer_list:
            raise NotAcceptable()

        for response_serializer in matching_response_serializer_list:
            if isinstance(response.response_body_input, response_serializer.get_response_data_type()):
                response.content, content_type = response_serializer.validate_and_serialize(
                    response.response_body_input, response_status_parameter
                )
                # Todo encoding from incoming accept charset
                response.text = response.content.decode()
                response._prepare_headers(content_type)
                return

        raise Response.ResponseBodyTypeNotSupportedException(
            f"Unable to find response data serializer for {http_accept} and response data type "
            f"{type(response.response_body_input)}"
        )



from restit.internal.http_accept import HttpAccept
from restit.internal.response_serializer_service import ResponseSerializerService
from restit.response import Response
from restit.rfc7807_http_problem import Rfc7807HttpProblem


class HttpExceptionResponseMaker:
    HTML_TEMPLATE = """<title>{code} {name}</title>
<h1>{name}</h1>
<p>{description}</p>
"""

    def __init__(self, rfc7807_http_problem: Rfc7807HttpProblem):
        self.rfc7807_http_problem = rfc7807_http_problem

    def create_response(self, http_accept: HttpAccept) -> Response:
        supported_media_types = [
            "text/html", "application/xhtml+xml", "application/json", "application/problem+json"
        ]
        best_match = http_accept.get_best_match(supported_media_types)
        if best_match is None:
            raise HttpExceptionResponseMaker.NoSupportedMIMETypesFoundException(http_accept.mime_types)

        best_match = best_match[0]

        if best_match in ["text/html", "application/xhtml+xml"]:
            response = self.create_html_response()
        elif best_match in ["application/json", "application/problem+json"]:
            response = self.create_rfc7807_json_response()
        else:
            response = self.create_plain_text_response()

        ResponseSerializerService.validate_and_serialize_response_body(response, http_accept, None)
        return response

    def create_html_response(self) -> Response:
        html_text = HttpExceptionResponseMaker.HTML_TEMPLATE.format(
            code=self.rfc7807_http_problem.status,
            name=self.rfc7807_http_problem.title,
            description=self.rfc7807_http_problem.detail
        )
        response = Response(
            response_body=html_text,
            status_code=self.rfc7807_http_problem.status,
            headers={"Content-Type": "text/html"}
        )
        return response

    def create_rfc7807_json_response(self) -> Response:
        """https://tools.ietf.org/html/rfc7807
        https://tools.ietf.org/html/rfc7807#section-3.1
        """
        response = Response(
            response_body=self.rfc7807_http_problem.to_json(),
            status_code=self.rfc7807_http_problem.status,
            headers={"Content-Type": "application/problem+json"}
        )
        return response

    def create_plain_text_response(self) -> Response:
        return Response(
            response_body=str(self.rfc7807_http_problem),
            status_code=self.rfc7807_http_problem.status
        )

    class NoSupportedMIMETypesFoundException(Exception):
        pass



from restit.common import get_open_api_resource_path
from restit.open_api.open_api_documentation import OpenApiDocumentation
from restit.request import Request
from restit.response import Response
from restit.static_directory_resource import StaticDirectoryResource


class OpenApiResource(StaticDirectoryResource):
    def __init__(self, open_api_documentation: OpenApiDocumentation):
        self.open_api_documentation = open_api_documentation
        super().__init__(open_api_documentation.path, get_open_api_resource_path())

    def get(self, request: Request, **path_params) -> Response:
        if path_params["file_name"] == "swagger.json":
            return Response(self.open_api_documentation.generate_spec())

        return super().get(request, **path_params)



import logging
import unittest

from marshmallow import Schema, fields

from restit import Resource, Request, Response, RestitTestApp, RestitApp, request_mapping
from restit.response_status_decorator import response_status


class MySchema(Schema):
    field1 = fields.String()
    field2 = fields.Integer()


@request_mapping("/")
class MyResource(Resource):

    @response_status(200, {"application/json": MySchema()}, "Everything is ok")
    def get(self, request: Request, **path_params) -> Response:
        request_body = request.get_request_body_as_type(dict)
        return Response({"field1": "Hans", "field2": "10"}, status_code=request_body["status"])


class ResponseStatusParameterTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.restit_test_app = RestitTestApp(RestitApp(resources=[MyResource()]))

    def test_status_supported(self):
        response = self.restit_test_app.get("/", json={"status": 200})
        self.assertEqual(200, response.get_status_code())
        self.assertEqual({'field1': 'Hans', 'field2': 10}, response.json())

    def test_status_unsupported(self):
        with self.assertLogs(level=logging.WARNING) as logs:
            response = self.restit_test_app.get("/", json={"status": 201})
            self.assertEqual(201, response.get_status_code())

            self.assertIn("WARNING:restit.resource:Response status code 201 is not expected for ", logs.output[0])



import os
from pathlib import Path
from typing import Union

from restit.internal.suffix_media_type_mapping import SUFFIX_MEDIA_TYPE_MAPPING
from restit.request import Request
from restit.resource import Resource
from restit.response import Response


class StaticDirectoryResource(Resource):
    def __init__(self, path: str, static_directory_path: Union[str, Path], entry_file: str = None):
        super().__init__()
        self.__request_mapping__ = path.rstrip("/") + r"/?(?P<file_name>\S*)"
        self.static_directory_path = static_directory_path
        self.entry_file = entry_file or "index.html"

    def get(self, request: Request, **path_params) -> Response:
        file_name = path_params["file_name"] or self.entry_file

        file_path = os.path.join(self.static_directory_path, file_name)
        with open(file_path, "rb") as fp:
            file_content = fp.read()

        _, suffix = os.path.splitext(file_name)
        content_type = SUFFIX_MEDIA_TYPE_MAPPING.get(suffix, )

        return Response(file_content, headers={"Content-Type": content_type})



class LicenseObject:
    def __init__(self, name: str, url: str):
        self.name = name
        self.url = url

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "url": self.url
        }



from http import HTTPStatus

import requests

from restit.request import Request
from restit.request_mapping_decorator import request_mapping
from restit.resource import Resource
from restit.response import Response
from test.base_test_server_test_case import BaseTestServerTestCase


@request_mapping("/")
class RequestBodyResource(Resource):
    def get(self, request: Request, **kwargs) -> Response:
        return Response({
            "form": dict(request.get_werkzeug_request().form),
            "data": request.get_werkzeug_request().data.decode(),
            "query_string": request.get_werkzeug_request().query_string.decode(),
            "query_parameters": request.get_query_parameters(),
            "content_type": request.get_werkzeug_request().content_type
        })

    def post(self, request: Request, **kwargs) -> Response:
        return Response({
            "form": dict(request.get_werkzeug_request().form),
            "data": request.get_werkzeug_request().data.decode(),
            "query_string": request.get_werkzeug_request().query_string.decode(),
            "query_parameters": request.get_query_parameters(),
            "content_type": request.get_werkzeug_request().content_type
        }, HTTPStatus.CREATED)


class RestitAppTestCase(BaseTestServerTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        BaseTestServerTestCase.resources = [RequestBodyResource()]
        BaseTestServerTestCase.setUpClass()

    def test_request_value_test_post_with_json(self):
        port = self.port
        response = requests.post(f"http://127.0.0.1:{port}/", json={"key": "value"})

        self.assertEqual(201, response.status_code)
        self.assertEqual({
            'data': '{"key": "value"}',
            'form': {},
            'query_parameters': {},
            'query_string': '',
            'content_type': 'application/json'
        }, response.json())

    def test_request_value_test_get_with_json(self):
        port = self.port
        response = requests.get(f"http://127.0.0.1:{port}/", json={"key": "value"})

        self.assertEqual(200, response.status_code)
        self.assertEqual({
            'data': '{"key": "value"}',
            'form': {},
            'query_parameters': {},
            'query_string': '',
            'content_type': 'application/json'
        }, response.json())

    def test_request_value_test_post_with_data(self):
        port = self.port
        response = requests.post(f"http://127.0.0.1:{port}/", data={"key": "value"})
        self.assertEqual(201, response.status_code)
        self.assertEqual({
            'data': '',
            'form': {'key': 'value'},
            'query_parameters': {},
            'query_string': '',
            'content_type': 'application/x-www-form-urlencoded'
        }, response.json())

    def test_request_value_test_get_with_data(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/", data={"key": "value"})
        self.assertEqual(200, response.status_code)
        self.assertEqual({
            'data': '',
            'form': {'key': 'value'},
            'query_parameters': {},
            'query_string': '',
            'content_type': 'application/x-www-form-urlencoded'
        }, response.json())

    def test_query_parameters_in_post(self):
        response = requests.post(f"http://127.0.0.1:{self.port}/?key=value")
        self.assertEqual(201, response.status_code)
        self.assertEqual({
            'data': '',
            'form': {},
            'query_parameters': {"key": "value"},
            'query_string': 'key=value',
            'content_type': 'text/plain'
        }, response.json())

    def test_query_parameters_in_get(self):
        response = requests.get(f"http://127.0.0.1:{self.port}/?key=value")
        self.assertEqual(200, response.status_code)
        self.assertEqual({
            'data': '',
            'form': {},
            'query_parameters': {"key": "value"},
            'query_string': 'key=value',
            'content_type': 'text/plain'
        }, response.json())



from marshmallow import Schema, fields

from restit import Resource, request_mapping, Response, Request, query_parameter
from restit.response_status_decorator import response_status
from test.restit.acceptence_test.todo_repo import TodoRepo


class TodosSchema(Schema):
    """A collection of todo names"""

    collection = fields.List(fields.String())


@request_mapping("/")
class TodosResource(Resource):
    def __init__(self, repo: TodoRepo):
        super().__init__()
        self.repo = repo

    @response_status(
        200, description="Get a list of all todo ids",
        content_types={"application/json": TodosSchema()}
    )
    @query_parameter(
        "sort", description="Flag if the todo list should be sorted", field_type=bool, default=False, required=False
    )
    def get(self, request: Request, **path_params) -> Response:
        """Get a list of all todo ids"""

        todo_ids = self.repo.get_todo_ids()

        return Response({"collection": todo_ids})



from typing import List, Type

from restit.common import create_dict_from_assignment_syntax, get_default_encoding
from restit.request_deserializer import RequestDeserializer


class DefaultFormDataDictDeserializer(RequestDeserializer):
    def get_content_type_list(self) -> List[str]:
        return ["application/x-www-form-urlencoded", "multipart/form-data", "application/x-url-encoded"]

    def deserialize(self, request_input: bytes, encoding: str = None) -> dict:
        return create_dict_from_assignment_syntax(request_input.decode(encoding or get_default_encoding()))

    def get_deserialized_python_type(self) -> Type:
        return dict



import os
from http import HTTPStatus
from pathlib import Path
from typing import Union

from restit.internal.suffix_media_type_mapping import SUFFIX_MEDIA_TYPE_MAPPING
from restit.response import Response


class StaticFileResponse(Response):
    def __init__(
            self, file_path: Union[str, Path],
            status_code: Union[int, HTTPStatus] = HTTPStatus.OK,
            headers: dict = None,
            suffix: str = None
    ):
        headers = headers or {}
        suffix = suffix or StaticFileResponse._get_suffix_from_file_path(file_path)
        content_type = SUFFIX_MEDIA_TYPE_MAPPING.get(suffix, )
        headers.setdefault("Content-Type", content_type)

        with open(file_path, "rb") as fp:
            file_content = fp.read()

        super().__init__(file_content, status_code, headers)

    @staticmethod
    def _get_suffix_from_file_path(file_path: str) -> str:
        _, suffix = os.path.splitext(file_path)
        return suffix



import unittest

from marshmallow import Schema, fields


class PythonDocTestCase(unittest.TestCase):
    def test_attribute_doc_test(self):
        class MySchema(Schema):
            "Can you access me?"
            field1 = fields.Integer()

            """And me?

            I have a bigger docstring!
            """
            field2 = fields.String()

        my_schema = MySchema()

        self.assertEqual("Can you access me?", my_schema.fields["field1"].parent.__doc__)
        # no ... parent is the class itself ... so no chance to access attribute docs :-(
        self.assertEqual("Can you access me?", my_schema.fields["field2"].parent.__doc__)



import logging
import signal
from contextlib import contextmanager
from threading import Thread
from time import sleep
from typing import Union, Callable, Generator, List
from wsgiref.simple_server import make_server, WSGIServer

LOGGER = logging.getLogger(__name__)


class DevelopmentServer:
    def __init__(self, wsgi_app: Callable, host: str = None, port: int = 0, stop_signals: List[int] = None):
        self.wsgi_app = wsgi_app
        self.host = host or "127.0.0.1"
        self.port = port
        self.stop_signals = stop_signals or [signal.SIGTERM, signal.SIGINT]
        self.server: Union[WSGIServer, None] = None
        self._thread: Union[Thread, None] = None
        self._is_running = False

    def start(self, blocking: bool = True) -> int:
        LOGGER.debug("Creating WSGI server for host %s and port %d", self.host, self.port)
        self._register_stop_signals()
        self.server = make_server(self.host, self.port, self.wsgi_app)
        self._thread = Thread(target=self.server.serve_forever)
        self._thread.start()
        self._is_running = True
        if blocking:
            LOGGER.info(
                "Starting development server in blocking mode at http://%s:%d/", self.host, self.server.server_port
            )
            self._wait_until_stopped()
        else:
            LOGGER.info("Development server is now running at http://%s:%d/", self.host, self.port)
            return self.server.server_port

    def _wait_until_stopped(self):
        while self._is_running:
            sleep(0.5)

    def _register_stop_signals(self):
        for stop_signal in self.stop_signals:
            LOGGER.debug("Registering signal %d as stop signal", stop_signal)
            signal.signal(stop_signal, self._stop_from_signal)

    def _stop_from_signal(self, signum: int, __):
        LOGGER.info("Received signal %d", signum)
        self.stop()

    def stop(self):
        LOGGER.info("Stopping development server")
        self._is_running = False
        self.server.shutdown()

    @contextmanager
    def start_in_context(self) -> Generator[int, None, None]:
        port = self.start(blocking=False)
        try:
            yield port
        finally:
            self.stop()



from setuptools import setup, find_packages

from restit import __version__

setup(
    name="restit",
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "marshmallow",
        "werkzeug"
    ],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



from functools import lru_cache
from typing import List, Type, Union, Any

from restit.internal.default_request_deserializer.default_any_bytes_deserializer import DefaultAnyBytesDeserializer
from restit.internal.default_request_deserializer.default_any_string_deserializer import DefaultAnyStringDeserializer
from restit.internal.default_request_deserializer.default_application_json_dict_deserializer import \
    DefaultApplicationJsonDictDeserializer
from restit.internal.default_request_deserializer.default_form_data_dict_deserializer import \
    DefaultFormDataDictDeserializer
from restit.request_deserializer import RequestDeserializer

_DEFAULT_REQUEST_DESERIALIZER = [
    DefaultApplicationJsonDictDeserializer(),
    DefaultFormDataDictDeserializer(),
    DefaultAnyBytesDeserializer(),
    DefaultAnyStringDeserializer()
]


class RequestDeserializerService:
    _REQUEST_DESERIALIZER: List[RequestDeserializer] = _DEFAULT_REQUEST_DESERIALIZER.copy()

    @staticmethod
    def register_request_deserializer(request_deserializer: RequestDeserializer):
        RequestDeserializerService._REQUEST_DESERIALIZER.insert(0, request_deserializer)

    @staticmethod
    def clear_all_request_deserializers():
        RequestDeserializerService._REQUEST_DESERIALIZER.clear()

    @staticmethod
    def restore_default_request_deserializers():
        RequestDeserializerService._REQUEST_DESERIALIZER = _DEFAULT_REQUEST_DESERIALIZER.copy()

    @staticmethod
    def deserialize_request_body(body: bytes, content_type: str, python_type: Type, encoding: str) -> Any:
        request_deserializer = RequestDeserializerService._find_deserializer(content_type, python_type)
        if request_deserializer is None:
            raise RequestDeserializerService.NoRequestDeserializerFoundException(
                f"Unable to find a request deserializer for content type {content_type} "
                f"to type {python_type}"
            )
        deserialized_value = request_deserializer.deserialize(body, encoding=encoding)
        return deserialized_value

    @staticmethod
    @lru_cache()
    def _find_deserializer(content_type: str, python_type: Type) -> Union[RequestDeserializer, None]:
        for request_deserializer in RequestDeserializerService._REQUEST_DESERIALIZER:
            if request_deserializer.can_handle(content_type, python_type):
                return request_deserializer

    class NoRequestDeserializerFoundException(Exception):
        pass



from typing import Dict, Union, Type

from marshmallow import Schema
from marshmallow.fields import Field
from werkzeug.exceptions import UnprocessableEntity, UnsupportedMediaType


class RequestBodyProperties:
    def __init__(
            self,
            content_types: Dict[str, Union[Schema, Field]],
            description: str,
            required: bool, validation_error_class=UnprocessableEntity
    ):
        self.content_types = content_types
        self.description = description
        self.required = required
        self.validation_error_class = validation_error_class
        self._check_request_properties_schema_type()

    def get_schema_for_content_type(self, content_type: str) -> Union[Schema, Type]:
        try:
            return self.content_types[content_type]
        except KeyError:
            raise UnsupportedMediaType()

    def _check_request_properties_schema_type(self):
        for schema_or_field in self.content_types.values():
            if not isinstance(schema_or_field, (Field, Schema)):
                raise RequestBodyProperties.UnsupportedSchemaTypeException(type(schema_or_field))

    class UnsupportedSchemaTypeException(Exception):
        pass



import unittest

from restit.request_mapping_decorator import PathIsNotStartingWithSlashException, request_mapping
from restit.resource import Resource


class CommonTestCase(unittest.TestCase):
    def test_request_mapping_is_not_starting_with_slash(self):
        with self.assertRaises(PathIsNotStartingWithSlashException):
            @request_mapping("not_starting_with_slash")
            class _(Resource):
                pass



import unittest

import requests

from restit.namespace import Namespace
from restit.request import Request
from restit.request_mapping_decorator import request_mapping
from restit.resource import Resource
from restit.response import Response
from restit.restit_app import RestitApp
from test.start_server_with_wsgi_app import start_server_with_wsgi_app


@request_mapping("/subpath")
class MyResource(Resource):
    def get(self, request: Request) -> Response:
        return Response("Hallo")


@request_mapping("/subpath2")
class MyResource2(Resource):
    def get(self, request: Request) -> Response:
        return Response("Hallo from subpath 2")


class NamespaceTestCase(unittest.TestCase):
    def test_something(self):
        namespace = Namespace("/huhu", resources=[MyResource()])
        namespace.register_resources([MyResource2()])

        restit_app = RestitApp(namespaces=[namespace])

        with start_server_with_wsgi_app(restit_app) as port:
            response = requests.get(f"http://127.0.0.1:{port}/huhu/subpath")
            self.assertEqual(200, response.status_code)
            self.assertEqual("Hallo", response.text)
            response = requests.get(f"http://127.0.0.1:{port}/huhu/subpath2")
            self.assertEqual(200, response.status_code)
            self.assertEqual("Hallo from subpath 2", response.text)



from typing import List, Tuple, Union

from restit.internal.mime_type import MIMEType


class HttpAccept:
    def __init__(self, mime_types: List[MIMEType]):
        self.mime_types: List[MIMEType] = sorted(mime_types, reverse=True)

    @staticmethod
    def from_accept_string(accept_string: str) -> "HttpAccept":
        return HttpAccept(
            [
                MIMEType.from_string(mime_type_string)
                for mime_type_string in accept_string.split(",")
            ]
        )

    def get_best_match(self, mime_type_strings: List[str]) -> Union[None, Tuple[str, MIMEType]]:
        for mime_type in self.mime_types:
            for mime_type_string in mime_type_strings:
                if mime_type.matches_mime_type_string(mime_type_string):
                    return mime_type_string, mime_type

    def __str__(self):
        return f"HttpAccept({[str(m) for m in self.mime_types]})"



import unittest

from restit.internal.resource_path import ResourcePath


class ResourcePathTestCase(unittest.TestCase):
    def test_match_no_path_param(self):
        resource_path = ResourcePath("/hello")

        self.assertEqual((True, {}), resource_path.get_match("/hello"))
        self.assertEqual((False, None), resource_path.get_match("/hello/"))
        self.assertEqual((False, None), resource_path.get_match("hello"))
        self.assertEqual((False, None), resource_path.get_match("/miau/hello"))

    def test_match_no_type(self):
        resource_path = ResourcePath("/hello/:id")

        self.assertEqual((True, {"id": "21"}), resource_path.get_match("/hello/21"))
        self.assertEqual((False, None), resource_path.get_match("/hello/"))
        self.assertEqual((False, None), resource_path.get_match("/hans/hello/21"))
        self.assertEqual((False, None), resource_path.get_match("/hello"))

    def test_match_multiple_path_params(self):
        resource_path = ResourcePath("/hello/:id/wuff/:miau")
        self.assertEqual((True, {'id': '21', 'miau': 'dog'}), resource_path.get_match("/hello/21/wuff/dog"))

    def test_match_path_param_type_int(self):
        resource_path = ResourcePath("/hello/:id<int>")
        self.assertEqual((True, {"id": 21}), resource_path.get_match("/hello/21"))
        self.assertEqual((False, None), resource_path.get_match("/hello/hans"))

        resource_path = ResourcePath("/hello/:id<integer>")
        self.assertEqual((True, {"id": 21}), resource_path.get_match("/hello/21"))
        self.assertEqual((False, None), resource_path.get_match("/hello/hans"))

    def test_match_path_param_type_string(self):
        resource_path = ResourcePath("/hello/:id<str>")
        self.assertEqual((True, {"id": "21"}), resource_path.get_match("/hello/21"))
        self.assertEqual((True, {"id": "hans"}), resource_path.get_match("/hello/hans"))

        resource_path = ResourcePath("/hello/:id<string>")
        self.assertEqual((True, {"id": "21"}), resource_path.get_match("/hello/21"))
        self.assertEqual((True, {"id": "hans"}), resource_path.get_match("/hello/hans"))

    def test_match_path_param_with_special_chars(self):
        resource_path = ResourcePath("/hello/:id_1<str>")
        self.assertEqual((True, {"id_1": "21"}), resource_path.get_match("/hello/21"))
        self.assertEqual((True, {"id_1": "hans"}), resource_path.get_match("/hello/hans"))

    def test_invalid_type_annotation(self):
        with self.assertRaises(ResourcePath.UnknownPathParamTypeAnnotation):
            ResourcePath("/hello/:id<invalid>")



from datetime import datetime

_TODOS = {
    1: ("Get beer", datetime(2020, 4, 1, 20, 0), "At least 10 boxes"),
    2: ("Get wine", datetime(2020, 4, 2, 20, 0), "At least 100 bottles"),
}


class TodoRepo:
    @staticmethod
    def get_todo_ids():
        return list(_TODOS.keys())

    @staticmethod
    def find_by_name(name: str):
        return _TODOS[name]



from marshmallow import fields, ValidationError


class BytesField(fields.Field):
    def _validate(self, value):
        if not isinstance(value, bytes):
            raise ValidationError('Invalid input type.')

        if value is None or value == b'':
            raise ValidationError('Invalid value')



import unittest

from restit.internal.http_accept import HttpAccept
from restit.internal.mime_type import MIMEType


class HttpAcceptTestCase(unittest.TestCase):
    def test_create_http_accept_from_string(self):
        http_accept_string = \
            'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,' \
            '*/*;q=0.8,application/signed-exchange;v=b3;q=0.9'

        http_accept = HttpAccept.from_accept_string(http_accept_string)
        self.assertEqual(7, len(http_accept.mime_types))
        self.assertEqual(MIMEType("text", "html"), http_accept.mime_types[0])
        self.assertEqual(MIMEType("application", "xhtml+xml"), http_accept.mime_types[1])
        self.assertEqual(MIMEType("image", "webp"), http_accept.mime_types[2])
        self.assertEqual(MIMEType("image", "apng"), http_accept.mime_types[3])
        self.assertEqual(MIMEType("application", "xml", 0.9, {"q": "0.9"}), http_accept.mime_types[4])
        self.assertEqual(MIMEType(
            "application", "signed-exchange", 0.9, {"v": "b3", "q": "0.9"}), http_accept.mime_types[5]
        )
        self.assertEqual(MIMEType("*", "*", 0.8, {"q": "0.8"}), http_accept.mime_types[6])

    def test_get_best_match(self):
        http_accept = HttpAccept([
            MIMEType("application", "json", 0.8),
            MIMEType("application", "xml", 0.9),
        ])

        self.assertEqual(
            ("application/xml", MIMEType("application", "xml", 0.9)),
            http_accept.get_best_match(["application/json", "application/xml"])
        )

    def test_get_best_match_wildcard(self):
        http_accept = HttpAccept([
            MIMEType("text", "*"),
            MIMEType("application", "xml"),
        ])

        self.assertEqual(
            ("text/html", MIMEType("text", "*")),
            http_accept.get_best_match(["text/html", "application/json"])
        )

    def test_get_best_match_wildcard_in_input(self):
        http_accept = HttpAccept([
            MIMEType("text", "html"),
            MIMEType("application", "xml", 0.9),
        ])

        self.assertEqual(
            ("text/*", MIMEType("text", "html")),
            http_accept.get_best_match(["text/*", "application/xml"])
        )

    def test_str_hook(self):
        http_accept = HttpAccept([MIMEType("text", "html")])

        self.assertEqual(
            "HttpAccept(['MIMEType(type=text, subtype=html, quality=1.0, details={})'])", str(http_accept)
        )



from restit.open_api.contact_object import ContactObject
from restit.open_api.license_object import LicenseObject


class InfoObject:
    def __init__(
            self,
            title: str,
            version: str,
            description: str = None,
            terms_of_service: str = None, contact: ContactObject = None, license: LicenseObject = None):
        self.title = title
        self.version = version
        self.description = description
        self.terms_of_service = terms_of_service
        self.contact = contact
        self.license = license

    def to_dict(self) -> dict:
        return {
            "title": self.title,
            "version": self.version,
            "description": self.description,
            "termsOfService": self.terms_of_service,
            "contact": self.contact.to_dict() if self.contact else None,
            "license": self.license.to_dict() if self.license else None
        }



import logging
from collections import namedtuple
from typing import List

from marshmallow.fields import Field, String

LOGGER = logging.getLogger(__name__)

QueryParameter = namedtuple("QueryParameter", ["name", "description", "field_type"])


# noinspection PyShadowingBuiltins
def query_parameter(name: str, description: str, field_type: Field = String):
    def decorator(func):
        _query_parameter = QueryParameter(name, description, field_type)
        registered_query_parameters: List[QueryParameter] = getattr(func, "__query_parameters__", [])
        LOGGER.debug(
            "Registering query parameter %s for %s", _query_parameter, func.__name__
        )
        registered_query_parameters.append(_query_parameter)
        setattr(func, "__query_parameters__", registered_query_parameters)
        return func

    return decorator



import re
from typing import Tuple, Union, Dict, AnyStr, Match


class ResourcePath:
    _TYPE_MAPPING = {
        "int": (r"\d+", int),
        "integer": (r"\d+", int),
        "string": (r"\S+", str),
        "str": (r"\S+", str),
        None: (r"\S+", str)
    }
    _PATH_PARAM_REGEX = re.compile(r":(\w+)(?:<(\w+)>)?")

    def __init__(self, request_mapping: str):
        self._request_mapping = request_mapping
        self._type_mapping = {}
        self._request_mapping_regex = self._transform_to_regex()

    def _transform_to_regex(self):
        regex_pattern = ResourcePath._PATH_PARAM_REGEX.sub(self._handle_path_param, self._request_mapping)
        regex_pattern = "^" + regex_pattern + "$"
        return re.compile(regex_pattern)

    def _handle_path_param(self, match: Match) -> str:
        self._type_mapping[match.group(1)] = \
            ResourcePath._TYPE_MAPPING.get(match.group(2), ResourcePath._TYPE_MAPPING["str"])[1]
        try:
            return f"(?P<{match.group(1)}>" + ResourcePath._TYPE_MAPPING[match.group(2)][0] + ")"
        except KeyError:
            raise ResourcePath.UnknownPathParamTypeAnnotation(match.group(2))

    def get_match(self, url: str) -> Tuple[bool, Union[None, Dict[str, AnyStr]]]:
        match = self._request_mapping_regex.match(url)
        if match:
            match_dict = {
                param: self._type_mapping.get(param, str)(value)
                for param, value in match.groupdict().items()
            }
            return True, match_dict
        return False, None

    @staticmethod
    def generate_url_with_path_parameter_values(request_mapping: str, path_param: dict) -> str:
        def _replace_with_path_param(match: Match):
            key = match.group(1)
            try:
                value = str(path_param[key])
            except KeyError:
                raise ResourcePath.ExpectedPathParameterForRequestMappingNotFoundException(
                    f"The path parameter {key} in request mapping '{request_mapping}' was not found in the provided "
                    f"path parameters {path_param}"
                )
            return value

        return ResourcePath._PATH_PARAM_REGEX.sub(_replace_with_path_param, request_mapping)

    class UnknownPathParamTypeAnnotation(Exception):
        pass

    class ExpectedPathParameterForRequestMappingNotFoundException(Exception):
        pass
